import csv
from pathlib import Path

from raccoon.utils.reporting import (
    generate_alignment_report,
    generate_combine_report,
    generate_mask_report,
    generate_phylo_report,
)


def _write_fasta(path: Path, records: list[tuple[str, str]]) -> None:
    with path.open("w") as handle:
        for rec_id, seq in records:
            handle.write(f">{rec_id}\n")
            for i in range(0, len(seq), 80):
                handle.write(seq[i:i + 80] + "\n")


def test_generate_combine_report_renders_template(tmp_path: Path) -> None:
    input_a = tmp_path / "a.fasta"
    input_b = tmp_path / "b.fasta"
    input_c = tmp_path / "c.fasta"
    _write_fasta(input_a, [("seqA|Loc1|2024-01-01", "ATGCGTNNNN")])
    _write_fasta(input_b, [("seqB|Loc2|2024-01-02", "ATGCGTATGC")])
    _write_fasta(input_c, [("seqC|Loc1|2024-01-03", "ATGC")])

    output_fasta = tmp_path / "combined.fasta"
    _write_fasta(output_fasta, [("seqA|Loc1|2024-01-01", "ATGCGTNNNN")])

    metadata_csv = tmp_path / "metadata.csv"
    with metadata_csv.open("w", newline="") as handle:
        writer = csv.writer(handle)
        writer.writerow(["id", "location", "date"])
        writer.writerow(["seqA", "Loc1", "2024-01-01"])
        writer.writerow(["seqB", "Loc2", "2024-01-02"])

    report_path = generate_combine_report(
        outdir=str(tmp_path),
        output_fasta=str(output_fasta),
        input_fastas=[str(input_a), str(input_b), str(input_c)],
        metadata_paths=[str(metadata_csv)],
        min_length=5,
        max_n_content=0.4,
    )

    html = Path(report_path).read_text()
    assert "Raccoon seq-qc report" in html
    assert "Inputs & sequences" in html or "Inputs &amp; sequences" in html
    assert "Report metadata" in html
    assert "metadata.csv" in html


def test_generate_alignment_report_renders_template(tmp_path: Path) -> None:
    alignment = tmp_path / "alignment.fasta"
    _write_fasta(alignment, [
        ("seq1", "ATGNNN"),
        ("seq2", "ATG---"),
    ])

    mask_csv = tmp_path / "mask.csv"
    with mask_csv.open("w", newline="") as handle:
        writer = csv.DictWriter(handle, fieldnames=["flagged", "type", "Minimum", "Maximum", "Length", "present_in", "note"])
        writer.writeheader()
        writer.writerow({
            "flagged": "2",
            "type": "site",
            "Minimum": "2",
            "Maximum": "2",
            "Length": "1",
            "present_in": "seq1,seq2",
            "note": "homoplasy",
        })

    report_path = generate_alignment_report(
        outdir=str(tmp_path),
        alignment_path=str(alignment),
        mask_file=str(mask_csv),
    )

    html = Path(report_path).read_text()
    assert "Raccoon aln-qc report" in html
    assert "Alignment N-content" in html
    assert "Flagged sites" in html


def test_generate_tree_report_renders_template(tmp_path: Path) -> None:
    treefile = tmp_path / "tree.nwk"
    treefile.write_text("(A|Loc1|2024-01-01:0.1,B|Loc2|2024-02-01:0.2);")

    flags_csv = tmp_path / "flags.csv"
    with flags_csv.open("w", newline="") as handle:
        writer = csv.DictWriter(handle, fieldnames=["mutation_type", "site", "mutation"])
        writer.writeheader()
        writer.writerow({"mutation_type": "convergent", "site": "100", "mutation": "A>G"})
        writer.writerow({"mutation_type": "reversion", "site": "200", "mutation": "G>A"})

    report_path = generate_phylo_report(
        outdir=str(tmp_path),
        treefile=str(treefile),
        flags_csv=str(flags_csv),
    )

    html = Path(report_path).read_text()
    assert "Raccoon tree-qc report" in html
    assert "Convergent mutations" in html
    assert "Reversions" in html


def test_generate_mask_report_renders_template(tmp_path: Path) -> None:
    alignment = tmp_path / "alignment.fasta"
    _write_fasta(alignment, [
        ("seq1", "ATGNNN"),
        ("seq2", "ATG---"),
    ])

    mask_csv = tmp_path / "mask.csv"
    with mask_csv.open("w", newline="") as handle:
        writer = csv.DictWriter(handle, fieldnames=["flagged", "type", "Minimum", "Maximum", "Length", "present_in", "note"])
        writer.writeheader()
        writer.writerow({
            "flagged": "2",
            "type": "site",
            "Minimum": "2",
            "Maximum": "2",
            "Length": "1",
            "present_in": "seq1",
            "note": "mask",
        })

    report_path = generate_mask_report(
        outdir=str(tmp_path),
        alignment_path=str(alignment),
        mask_file=str(mask_csv),
        output_alignment=str(tmp_path / "alignment.masked.fasta"),
    )

    html = Path(report_path).read_text()
    assert "Raccoon mask report" in html
    assert "Masked sites by sequence" in html
    assert "Mask file entries" in html


#!/usr/bin/env python3
"""Command-line interface for raccoon toolkit."""

import sys
import os
import argparse
import logging
from raccoon import __version__, _program
from raccoon.commands import alignment as alignment_cmd, phylo as phylo_cmd, combine as combine_cmd, mask as mask_cmd


def build_parser():
    parser = argparse.ArgumentParser(prog=_program, description="raccoon toolkit")
    parser.add_argument('-v','--version', action='version', version=f'raccoon {__version__}', help='Show version and exit')
    parser.add_argument('-V','--verbose', action='count', default=0, help='Increase verbosity (can specify multiple times)')
    sub = parser.add_subparsers(dest='command')
    sub.required = True

    c = sub.add_parser('seq-qc', help='sequence QC (combine and harmonise FASTA files)')
    c.add_argument('inputs', nargs='+', help='Input FASTA files (one or more)')
    c.add_argument('-o', '--output', default='combined.fasta', help='Output FASTA file (use - for stdout)')
    c.add_argument('--metadata', nargs='+', action='extend', default=None, help='Metadata CSV(s) for harmonised headers')
    c.add_argument('--metadata-delimiter', default=',', help='Metadata delimiter (default: ,; auto-detects .tsv/.tab)')
    c.add_argument('--metadata-id-field', default='id', help='Metadata id column (default: id)')
    c.add_argument('--metadata-location-field', default='location', help='Metadata location column (default: location)')
    c.add_argument('--metadata-date-field', default='date', help='Metadata date column (default: date)')
    c.add_argument('--header-separator', default='|', help='Header separator (default: |)')
    c.add_argument('--id-delimiter', default='|', help='Delimiter for parsing IDs from input headers (default: |)')
    c.add_argument('--id-field', type=int, default=0, help='0-based field index for ID parsing (default: 0)')
    c.add_argument('--min-length', type=int, default=None, help='Minimum sequence length to keep')
    c.add_argument('--max-n-content', type=float, default=None, help='Maximum N content proportion to keep (e.g. 0.1)')
    c.set_defaults(func=combine_cmd.main)

    a = sub.add_parser('aln-qc', help='alignment QC')
    a.add_argument('alignment', help='Input alignment fasta file')
    a.add_argument('-t','--sequence-type', choices=['nt','aa'], default='nt', dest='sequence_type', help='Sequence type (default: nt)')
    a.add_argument('-d','--output-dir', default='.', dest='output_dir', help='Output directory')
    a.add_argument('--genbank', help='GenBank file for frame-breaking indel checks', dest='genbank', default=None)
    a.add_argument('--reference-id', help='Reference sequence ID in alignment (for GenBank features)', dest='reference_id', default=None)
    a.add_argument('--n-threshold', type=float, default=0.2, dest='n_threshold', help='N content threshold for flagging (default: 0.2)')
    a.add_argument('--cluster-window', type=int, default=10, dest='cluster_window', help='Window size for clustered SNP detection (default: 10bp)')
    a.add_argument('--cluster-count', type=int, default=3, dest='cluster_count', help='Min SNPs in window to flag as clustered (default: 3)')
    a.add_argument('--mask-clustered', default=True, action=argparse.BooleanOptionalAction, dest='mask_clustered', help='Mask clustered SNPs (default: True)')
    a.add_argument('--mask-n-adjacent', default=True, action=argparse.BooleanOptionalAction, dest='mask_n_adjacent', help='Mask SNPs adjacent to Ns (default: True)')
    a.add_argument('--mask-gap-adjacent', default=True, action=argparse.BooleanOptionalAction, dest='mask_gap_adjacent', help='Mask SNPs adjacent to gaps (default: True)')
    a.add_argument('--mask-frame-break', default=True, action=argparse.BooleanOptionalAction, dest='mask_frame_break', help='Mask frame-breaking indels (default: True)')
    a.set_defaults(func=alignment_cmd.main)

    m = sub.add_parser('mask', help='apply a mask file to an alignment')
    m.add_argument('alignment', help='Input alignment fasta file')
    m.add_argument('--mask-file', required=True, dest='mask_file', help='Mask CSV file from aln-qc')
    m.add_argument('-o', '--output', dest='output', default=None, help='Output masked alignment file')
    m.add_argument('-d', '--outdir', dest='outdir', default='.', help='Output directory')
    m.add_argument('-t', '--sequence-type', choices=['nt','aa'], default='nt', dest='sequence_type', help='Sequence type (default: nt)')
    m.set_defaults(func=mask_cmd.main)

    p = sub.add_parser('tree-qc', help='phylogenetic QC')
    p.add_argument('--assembly-refs', help='Assembly references fasta', dest='assembly_refs', default=None)
    p.add_argument('--phylogeny', help='Phylogeny tree file or base name', dest='phylogeny', required=True)
    p.add_argument('--outdir','-d', help='Output directory', dest='outdir', default='.')
    p.add_argument('--mask-file', help='Mask output file', dest='mask_file', default=None)
    p.add_argument('--outgroup-ids', help='Comma-separated outgroup ids', dest='outgroup_ids', default=None)
    p.add_argument('--alignment', help='Alignment fasta used with ASR state file', dest='alignment', default=None)
    p.add_argument('--asr-state', help='Ancestral state reconstruction file in the format output by IQTREE', dest='asr_state', default=None)
    p.add_argument('--tree-format', choices=['auto','newick','nexus'], default='auto', dest='tree_format', help='Tree format (default: auto)')
    p.add_argument('--long-branch-sd', type=float, default=3.0, dest='long_branch_sd', help='Std dev threshold for long branch flagging (default: 3.0)')
    p.add_argument('--adar-window', type=int, default=300, dest='adar_window', help='Max distance (bp) for ADAR cluster window (default: 300)')
    p.add_argument('--adar-min-count', type=int, default=3, dest='adar_min_count', help='Min ADAR sites in window to flag a branch (default: 3)')
    p.add_argument('--run-apobec', action='store_true', dest='run_apobec', help='Run APOBEC3 phylo checks')
    p.add_argument('--run-adar', action='store_true', dest='run_adar', help='Run ADAR phylo checks')
    p.add_argument('--height', help='Figure height', dest='height', default=None)
    p.set_defaults(func=phylo_cmd.main)

    return parser


def main(sysargs=None):
    if sysargs is None:
        sysargs = sys.argv[1:]
    parser = build_parser()
    args = parser.parse_args(sysargs)

    # configure logging
    level = logging.WARNING
    if getattr(args,'verbose',0) >=2:
        level = logging.DEBUG
    elif getattr(args,'verbose',0) == 1:
        level = logging.INFO
    logging.basicConfig(level=level, format='%(levelname)s: %(message)s')

    try:
        return args.func(args)
    except Exception as e:
        logging.exception("Error running raccoon")
        return 2


if __name__ == '__main__':
    sys.exit(main())


import csv
import sys
from pathlib import Path

import pytest

REPO_ROOT = Path(__file__).resolve().parents[1]
if str(REPO_ROOT) not in sys.path:
    sys.path.insert(0, str(REPO_ROOT))

from raccoon.utils import plotly_baltic


class FakeNode:
    def __init__(self, name, x, y, branch_type="leaf", children=None):
        self.name = name
        self.x = x
        self.y = y
        self.branchType = branch_type
        self.children = children or []
        self.traits = {}

    def is_leaflike(self):
        return self.branchType == "leaf"

    def is_node(self):
        return self.branchType == "internal"


class FakeTree:
    def __init__(self):
        self.root = FakeNode("root", x=0.0, y=1.5, branch_type="internal")
        self.tip_a = FakeNode("tipA", x=1.0, y=1.0, branch_type="leaf")
        self.tip_b = FakeNode("tipB", x=1.1, y=2.0, branch_type="leaf")
        self.root.children = [self.tip_a, self.tip_b]
        self.Objects = [self.root, self.tip_a, self.tip_b]

    def drawTree(self):
        return None

    def getInternal(self):
        return [self.root]


def _write_branch_snps_csv(path: Path) -> None:
    rows = [
        {"parent": "root", "child": "tipA", "site": "10", "snp": "G->A", "dimer": "GA"},
        {"parent": "root", "child": "tipB", "site": "20", "snp": "C->T", "dimer": "TC"},
    ]
    with path.open("w", newline="") as handle:
        writer = csv.DictWriter(handle, fieldnames=["parent", "child", "site", "snp", "dimer"])
        writer.writeheader()
        for row in rows:
            writer.writerow(row)


@pytest.fixture(autouse=True)
def patch_tree(monkeypatch):
    monkeypatch.setattr(plotly_baltic, "load_tree", lambda *_args, **_kwargs: FakeTree())
    monkeypatch.setattr(plotly_baltic, "ensure_node_label", lambda node: node.name)
    yield


def test_build_tree_plot_includes_branch_mutations_and_labels(tmp_path):
    csv_path = tmp_path / "branch_snps.csv"
    _write_branch_snps_csv(csv_path)

    html = plotly_baltic.build_tree_plot("tree", branch_snps_path=str(csv_path))

    assert "Branch: root_tipA" in html
    assert ("10: G->A (GA)" in html) or ("10: G-\\u003eA (GA)" in html)
    assert "tipA" in html
    assert "Reset view" in html
    assert "Expand tree" in html


def test_build_tree_plot_without_branch_snps(tmp_path):
    html = plotly_baltic.build_tree_plot("tree", branch_snps_path=str(tmp_path / "missing.csv"))
    assert "Reset view" in html
    assert "Branch:" not in html



import os
import tempfile
import pytest
from raccoon.commands import alignment


class MockArgs:
    def __init__(self, **kwargs):
        self.alignment = kwargs.get('alignment', 'test.fasta')
        self.output_dir = kwargs.get('output_dir', '.')
        self.genbank = kwargs.get('genbank', None)
        self.reference_id = kwargs.get('reference_id', None)
        self.n_threshold = kwargs.get('n_threshold', 0.2)
        self.cluster_window = kwargs.get('cluster_window', 10)
        self.cluster_count = kwargs.get('cluster_count', 3)


def test_output_directory_creation(tmp_path):
    """Test that output directory is created if it doesn't exist."""
    new_dir = tmp_path / 'new_output_dir'
    assert not new_dir.exists()

    args = MockArgs(alignment='nonexistent.fasta', output_dir=str(new_dir))
    # will fail due to missing alignment file, but directory should be created first
    result = alignment.main(args)
    
    # directory should have been created
    assert new_dir.exists()
    assert new_dir.is_dir()


def test_output_directory_writable_check(tmp_path):
    """Test that an error is raised if output directory is not writable."""
    readonly_dir = tmp_path / 'readonly'
    readonly_dir.mkdir()
    
    # make it read-only
    os.chmod(str(readonly_dir), 0o444)
    
    args = MockArgs(alignment='nonexistent.fasta', output_dir=str(readonly_dir))
    result = alignment.main(args)
    
    # should fail with write permission error
    assert result == 1
    
    # restore permissions for cleanup
    os.chmod(str(readonly_dir), 0o755)


def test_output_directory_defaults_to_cwd(tmp_path, monkeypatch, caplog):
    """Test that output_dir defaults to current working directory if not provided."""
    monkeypatch.chdir(str(tmp_path))
    
    args = MockArgs(alignment='nonexistent.fasta', output_dir=None)
    # will fail due to missing alignment, but should use cwd
    result = alignment.main(args)
    
    # should have logged the error at least
    assert result != 0  # should fail



"""Plotting helpers for raccoon.

These functions take parsed data (or file paths) and create figures. They are
pure (accept an optional Matplotlib `ax`) and suitable for headless testing.
"""
from typing import Optional
import logging
import math
import matplotlib as mpl
import matplotlib.pyplot as plt

# Try to import baltic lazily when plotting functions are called
try:
    import baltic as bt
except Exception:
    bt = None

plt.switch_backend('Agg')

mpl.rcParams.update({'font.size': 18})
new_rc_params = {'text.usetex': False, 'svg.fonttype': 'none'}
mpl.rcParams.update(new_rc_params)
plt.rcParams['font.family'] = 'Helvetica'


def _read_branch_snps(branch_snps):
    """Read branch snps CSV into a dict if a filepath provided."""
    import csv
    import collections

    if isinstance(branch_snps, dict):
        return branch_snps

    d = collections.defaultdict(list)
    with open(branch_snps, 'r') as f:
        reader = csv.DictReader(f)
        for row in reader:
            branch = f"{row['parent']}_{row['child']}"
            d[branch].append((row['site'], row['snp'], row['dimer']))
    return d


def make_reversion_tree_figure(outfile: str, branch_snps, branch_reversions, will_be_reverted, treefile: str, w: Optional[float], h: Optional[float], ax: Optional[plt.Axes] = None):
    """Plot tree figure showing reversions.

    `branch_snps` can be either a path to a CSV or a pre-parsed dict of branch->snps.
    """
    if bt is None:
        raise RuntimeError('baltic is required for plotting trees')

    branch_snps_dict = _read_branch_snps(branch_snps)

    my_tree = bt.loadNewick(treefile, absoluteTime=False)

    # heuristic sizing
    r2t = 200000 * my_tree.treeHeight
    if w is None:
        width = int(math.sqrt(r2t) * 3) if r2t < 200 else 25
    else:
        width = float(w)
    if h is None:
        height = int(math.sqrt(my_tree.ySpan) * 2) if my_tree.ySpan < 300 else 40
    else:
        height = float(h)

    if ax is None:
        fig, ax = plt.subplots(figsize=(width, height), facecolor='w')
    else:
        fig = ax.figure

    increment = my_tree.treeHeight / 150

    x_attr = lambda k: k.height
    su_func = lambda k: 50 - 30 * k.height / my_tree.treeHeight
    s_func = lambda k: 50 - 20 * k.height / my_tree.treeHeight
    c_func = lambda k: 'dimgrey'

    my_tree.plotTree(ax, x_attr=x_attr)
    my_tree.plotPoints(ax, size=s_func, colour=c_func, x_attr=x_attr)
    mpl.rcParams['font.family'] = 'sans-serif'

    text_x_attr = lambda k: k.height + (increment * 4)
    my_tree.addText(ax, x_attr=text_x_attr, target=lambda k: k.is_leaf(), text=lambda k: k.name)

    for k in my_tree.Objects:
        current_node = k
        if k.branchType == 'leaf':
            current_node.traits['label'] = k.name

        try:
            parent_name = current_node.parent.traits['label']
        except Exception:
            continue
        node_name = current_node.traits['label']
        branch_name = f"{parent_name}_{node_name}"

        if branch_name in branch_snps_dict:
            snps = []
            reversions = []
            tb_reversions = []
            if branch_name in branch_reversions:
                for s in branch_reversions[branch_name]:
                    reversions.append(s)
            if branch_name in will_be_reverted:
                for s in will_be_reverted[branch_name]:
                    tb_reversions.append(s)

            snp_placement = current_node.parent.height + increment
            rev_placement = (current_node.parent.height + current_node.height) / 2
            tb_rev_placement = (current_node.parent.height + current_node.height) / 2

            for s in branch_snps_dict[branch_name]:
                site, snp, dimer = s
                if snp == 'G->A':
                    if dimer in ['GA']:
                        snps.append((1, '#995e62'))
                    else:
                        snps.append((2, '#d9b660'))
                elif snp == 'C->T':
                    if dimer in ['TC']:
                        snps.append((1, '#995e62'))
                    else:
                        snps.append((2, '#d9b660'))
                else:
                    snps.append((2, '#d9b660'))

            for reversion in reversions:
                ax.text(s=reversion, x=rev_placement, y=k.y + 0.5, rotation='vertical')
                ax.scatter([rev_placement], [k.y], color='black', s=150, marker=8)
                rev_placement += 3 * increment
            for tb_reversion in tb_reversions:
                ax.text(s=tb_reversion, x=tb_rev_placement, y=k.y + 0.5, rotation='vertical')
                ax.scatter([tb_rev_placement], [k.y], color='black', s=150, marker=9)
                tb_rev_placement += 3 * increment

            for snp in sorted(snps, key=lambda x: x[0]):
                ax.scatter([snp_placement], [k.y + 0.5], color=snp[1], s=30)
                snp_placement += increment

    [ax.spines[loc].set_visible(False) for loc in ['top', 'right', 'left', 'bottom']]
    ax.tick_params(axis='y', size=0)
    ax.tick_params(axis='x', size=0)

    ax.set_yticklabels([])
    ax.set_xticklabels([])

    plt.savefig(f"{outfile}.svg", bbox_inches='tight')
    plt.savefig(f"{outfile}.png", bbox_inches='tight', transparent=True)


def make_convergence_tree_figure(outfile: str, branch_snps, branch_convergence, treefile: str, w: Optional[float], h: Optional[float], ax: Optional[plt.Axes] = None):
    """Plot tree figure showing convergent SNPs."""
    if bt is None:
        raise RuntimeError('baltic is required for plotting trees')

    branch_snps_dict = _read_branch_snps(branch_snps)

    my_tree = bt.loadNewick(treefile, absoluteTime=False)

    r2t = 200000 * my_tree.treeHeight
    if w is None:
        width = int(math.sqrt(r2t) * 3) if r2t < 200 else 25
    else:
        width = float(w)
    if h is None:
        height = int(math.sqrt(my_tree.ySpan) * 2) if my_tree.ySpan < 300 else 40
    else:
        height = float(h)

    if ax is None:
        fig, ax = plt.subplots(figsize=(width, height), facecolor='w')
    else:
        fig = ax.figure

    increment = my_tree.treeHeight / 150

    x_attr = lambda k: k.height
    s_func = lambda k: 50 - 20 * k.height / my_tree.treeHeight
    c_func = lambda k: 'dimgrey'

    my_tree.plotTree(ax, x_attr=x_attr)
    my_tree.plotPoints(ax, size=s_func, colour=c_func, x_attr=x_attr)
    mpl.rcParams['font.family'] = 'sans-serif'

    text_x_attr = lambda k: k.height + (increment * 4)
    my_tree.addText(ax, x_attr=text_x_attr, target=lambda k: k.is_leaf(), text=lambda k: k.name)

    for k in my_tree.Objects:
        current_node = k
        if k.branchType == 'leaf':
            current_node.traits['label'] = k.name

        try:
            parent_name = current_node.parent.traits['label']
        except Exception:
            continue
        node_name = current_node.traits['label']
        branch_name = f"{parent_name}_{node_name}"

        if branch_name in branch_snps_dict:
            snps = []
            convergent_snps = []
            if branch_name in branch_convergence:
                for s in branch_convergence[branch_name]:
                    convergent_snps.append(s)

            snp_placement = current_node.parent.height + increment
            c_placement = (current_node.parent.height + current_node.height) / 2
            for s in branch_snps_dict[branch_name]:
                site, snp, dimer = s
                if snp == 'G->A':
                    if dimer in ['GA']:
                        snps.append((1, '#995e62'))
                    else:
                        snps.append((2, '#d9b660'))
                elif snp == 'C->T':
                    if dimer in ['TC']:
                        snps.append((1, '#995e62'))
                    else:
                        snps.append((2, '#d9b660'))
                else:
                    snps.append((2, '#d9b660'))

            for c_snp in convergent_snps:
                ax.text(s=c_snp, x=c_placement, y=k.y + 0.5, rotation='vertical')
                ax.scatter([c_placement], [k.y], color='black', s=150, marker='d')
                c_placement += 2 * increment

            for snp in sorted(snps, key=lambda x: x[0]):
                ax.scatter([snp_placement], [k.y + 0.5], color=snp[1], s=30)
                snp_placement += increment

    [ax.spines[loc].set_visible(False) for loc in ['top', 'right', 'left', 'bottom']]
    ax.tick_params(axis='y', size=0)
    ax.tick_params(axis='x', size=0)

    ax.set_yticklabels([])
    ax.set_xticklabels([])

    plt.savefig(f"{outfile}.svg", bbox_inches='tight')
    plt.savefig(f"{outfile}.png", bbox_inches='tight', transparent=True)


def make_reconstruction_tree_figure_w_labels(outfile: str, branch_snps, treefile: str, point_style: str, justification: str, w: Optional[float] = None, h: Optional[float] = None, ax: Optional[plt.Axes] = None):
    """Plot reconstruction tree figure with labelled SNPs."""
    if bt is None:
        raise RuntimeError('baltic is required for plotting trees')

    branch_snps_dict = _read_branch_snps(branch_snps)

    my_tree = bt.loadNewick(treefile, absoluteTime=False)

    if w is None:
        r2t = 200000 * my_tree.treeHeight
        width = int(math.sqrt(r2t) * 3) if r2t < 200 else 25
    else:
        width = int(w)

    if h is None:
        height = int(math.sqrt(my_tree.ySpan) * 2) if my_tree.ySpan < 300 else 40
    else:
        height = int(h)

    if ax is None:
        fig, ax = plt.subplots(figsize=(width, height), facecolor='w')
    else:
        fig = ax.figure

    x_attr = lambda k: k.height
    s_func = lambda k: 50 - 20 * k.height / my_tree.treeHeight
    c_func = lambda k: 'dimgrey'

    my_tree.plotTree(ax, x_attr=x_attr)
    my_tree.plotPoints(ax, size=s_func, colour=c_func, x_attr=x_attr)
    mpl.rcParams['font.family'] = 'sans-serif'

    text_x_attr = lambda k: k.height + (my_tree.treeHeight / 150 * 4)
    my_tree.addText(ax, x_attr=text_x_attr, target=lambda k: k.is_leaf(), text=lambda k: k.name)

    right_settings = {"apobec": (1, "#995E62"), "non_apobec": (2, "#D9B660")}
    left_settings = {"apobec": (2, "#995E62"), "non_apobec": (1, "#D9B660")}

    for k in my_tree.Objects:
        if k.branchType == 'leaf':
            k.traits['label'] = k.name

        node_name = k.traits.get('label')
        if not node_name:
            continue
        
        try:
            parent_name = k.parent.traits.get('label')
            if not parent_name:
                continue
        except Exception:
            continue
        branch_name = f"{parent_name}_{node_name}"

        if branch_name in branch_snps_dict:
            if justification == 'right':
                setting_dict = right_settings
                snp_placement = k.height - my_tree.treeHeight / 150
            else:
                setting_dict = left_settings
                snp_placement = k.parent.height + (my_tree.treeHeight / 150) / 2

            snps = []
            for s in branch_snps_dict[branch_name]:
                site, snp, dimer = s
                if snp == 'G->A':
                    if dimer in ['GA']:
                        snps.append(setting_dict['apobec'])
                    else:
                        snps.append(setting_dict['non_apobec'])
                elif snp == 'C->T':
                    if dimer in ['TC']:
                        snps.append(setting_dict['apobec'])
                    else:
                        snps.append(setting_dict['non_apobec'])
                else:
                    snps.append(setting_dict['non_apobec'])

            for snp in sorted(snps, key=lambda x: x[0]):
                if point_style == 'circle':
                    ax.scatter([snp_placement], [k.y + 0.5], color=snp[1], s=50)
                else:
                    import matplotlib.patches as patches

                    rect = patches.Rectangle((snp_placement, k.y - 0.5), my_tree.treeHeight / 150 / 2, 1, alpha=1, fill=True, edgecolor='none', facecolor=snp[1])
                    ax.add_patch(rect)

                if justification == 'right':
                    snp_placement -= my_tree.treeHeight / 150
                else:
                    snp_placement += my_tree.treeHeight / 150

    [ax.spines[loc].set_visible(False) for loc in ['top', 'right', 'left']]
    ax.tick_params(axis='y', size=0)
    ax.tick_params(axis='x', size=0)

    ax.set_yticklabels([])
    ax.set_xticklabels([])

    plt.margins(0.1, 0.1, tight=True)
    plt.savefig(f"{outfile}.svg", bbox_inches='tight')
    plt.savefig(f"{outfile}.png", bbox_inches='tight')


"""
Raccoon: Virus Genome Alignment QC Curation Package

A comprehensive toolkit for curating virus genome and sub-genomic sequence alignments
and flagging various QC issues including clustered SNPs, convergent SNPs, ADAR edits,
and APOBEC3 edits.
"""

_program = "raccoon"
__version__ = "1.0.0"



"""I/O validation and file handling utilities."""
import os
import logging
from typing import Iterable

from Bio import SeqIO
import sys
import tempfile
from typing import Tuple, Optional


def ensure_output_directory(outdir: str) -> bool:
    """Ensure output directory exists and is writable.
    
    Creates the directory if it doesn't exist. Returns True on success,
    logs error and returns False on failure.
    """
    if not os.path.isdir(outdir):
        try:
            os.makedirs(outdir, exist_ok=True)
            logging.info(f"Created output directory: {outdir}")
        except OSError as exc:
            logging.error(f"Failed to create output directory '{outdir}': {exc}")
            return False
    
    if not os.access(outdir, os.W_OK):
        logging.error(f"Output directory '{outdir}' is not writable")
        return False
    
    return True


def ensure_parent_directory(path: str) -> bool:
    """Ensure the parent directory for a file path exists and is writable."""
    parent = os.path.dirname(path) or "."
    return ensure_output_directory(parent)


def validate_input_file(filepath: str, name: str = "input file") -> bool:
    """Validate that an input file exists and is readable.
    
    Returns True if file is valid, logs error and returns False otherwise.
    """
    if not filepath:
        return True  # optional files are OK
    
    if not os.path.isfile(filepath):
        logging.error(f"{name} '{filepath}' does not exist")
        return False
    
    if not os.access(filepath, os.R_OK):
        logging.error(f"{name} '{filepath}' is not readable")
        return False
    
    return True


def validate_alignment_file(filepath: str) -> bool:
    """Validate that an alignment file exists and is readable."""
    # support stdin (use '-' to indicate reading from STDIN)
    if filepath == '-' or filepath == '':
        # cannot fully validate stdin here (consuming it would prevent later reads)
        # assume caller will handle reading from stdin into a temp file
        return True

    if not os.path.isfile(filepath):
        logging.error(f"Alignment file '{filepath}' does not exist")
        return False
    
    if not os.access(filepath, os.R_OK):
        logging.error(f"Alignment file '{filepath}' is not readable")
        return False
    
    # ensure sequences in the alignment are consistent lengths (requires Biopython)
    try:
        lengths = set()
        for rec in SeqIO.parse(filepath, "fasta"):
            lengths.add(len(rec.seq))

        if not lengths:
            logging.error(f"Alignment file '{filepath}' contains no sequences")
            return False

        if len(lengths) != 1:
            logging.error(f"Alignment file '{filepath}' contains sequences of differing lengths: {sorted(lengths)}")
            return False

    except Exception as exc:  # pragma: no cover - defensive
        logging.error(f"Failed to read alignment file '{filepath}': {exc}")
        return False

    return True


def validate_genbank_file(filepath: str) -> bool:
    """Validate that a GenBank file exists and is readable (if provided)."""
    if not filepath:
        return True  # optional
    return validate_input_file(filepath, "GenBank file")


def validate_reference_file(filepath: str, name: str = "Reference file") -> bool:
    """Validate that a reference file exists and is readable (if provided)."""
    if not filepath:
        return True  # optional
    return validate_input_file(filepath, name)


def prepare_alignment_input(filepath: str) -> Tuple[str, Optional[str]]:
    """Prepare an alignment input path for downstream processing.

    If `filepath` is '-' the function reads stdin and writes it into a temporary
    FASTA file, returning the temp path and the temp path as the second element.
    For normal files it returns the original path and None.
    Caller is responsible for calling `cleanup_temp_file` on the returned
    temp path when finished.
    """
    if filepath == '-' or filepath == '':
        tmp = tempfile.NamedTemporaryFile(delete=False, suffix='.fasta')
        try:
            data = sys.stdin.buffer.read()
            tmp.write(data)
        finally:
            tmp.close()
        return tmp.name, tmp.name
    return filepath, None


def cleanup_temp_file(temp_path: Optional[str]) -> None:
    """Remove a temporary file created by `prepare_alignment_input`.

    Silently ignores missing files and errors.
    """
    if not temp_path:
        return
    try:
        if os.path.exists(temp_path):
            os.unlink(temp_path)
    except Exception:
        logging.debug("Failed to remove temp file %s", temp_path)


def resolve_existing_file(filepath: str, outdir: Optional[str], name: str) -> Optional[str]:
    """Resolve a file path directly or relative to outdir.

    Returns the resolved path if readable, otherwise None.
    """
    if not filepath:
        logging.error("%s is required", name)
        return None

    if validate_input_file(filepath, name):
        return filepath

    if outdir:
        candidate = os.path.join(outdir, filepath)
        if validate_input_file(candidate, name):
            return candidate

    return None


def resolve_asr_state_file(state_file: Optional[str], treefile: Optional[str]) -> Optional[str]:
    """Resolve ASR state file path (explicit or based on treefile)."""
    if state_file:
        if validate_input_file(state_file, "ASR state file"):
            return state_file
        return None

    if not treefile:
        return None

    candidates = [f"{treefile}.state", f"{os.path.splitext(treefile)[0]}.state"]
    for candidate in candidates:
        if os.path.isfile(candidate) and validate_input_file(candidate, "ASR state file"):
            return candidate

    return None



import csv

from Bio import AlignIO
from Bio.Align import MultipleSeqAlignment
from Bio.Seq import Seq
from Bio.SeqFeature import FeatureLocation, SeqFeature
from Bio.SeqRecord import SeqRecord

from raccoon.commands import alignment


def _write_genbank(path, record):
    record.annotations["molecule_type"] = "DNA"
    with open(path, "w") as handle:
        from Bio import SeqIO
        SeqIO.write(record, handle, "genbank")


def _build_alignment(tmp_path):
    # reference sequence and CDS covering full length
    ref_seq = "ATG" * 10
    ref_record = SeqRecord(Seq(ref_seq), id="ref", name="ref")
    ref_record.features.append(SeqFeature(FeatureLocation(0, len(ref_seq)), type="CDS"))

    # baseline identical sequence
    s1 = SeqRecord(Seq(ref_seq), id="s1")

    # clustered SNPs within 10bp window (positions 4, 6, 8 are unique)
    s_cluster_list = list(ref_seq)
    for pos, nt in [(4, "C"), (6, "G"), (8, "T")]:
        s_cluster_list[pos] = nt
    s_cluster = SeqRecord(Seq("".join(s_cluster_list)), id="s_cluster")

    # SNP next to N plus some Ns to trigger high-N sequence
    s_n_list = list(ref_seq)
    s_n_list[12] = "G"
    s_n_list[13] = "N"
    s_n_list[14] = "N"
    s_n = SeqRecord(Seq("".join(s_n_list)), id="s_n")

    # frame-breaking indel (single gap within CDS)
    s_frame_list = list(ref_seq)
    s_frame_list[5] = "-"
    s_frame = SeqRecord(Seq("".join(s_frame_list)), id="s_frame")

    aln = MultipleSeqAlignment([ref_record, s1, s_cluster, s_n, s_frame])

    aln_path = tmp_path / "test_alignment.fasta"
    AlignIO.write(aln, str(aln_path), "fasta")

    gb_path = tmp_path / "ref.gb"
    _write_genbank(gb_path, ref_record)

    return aln_path, gb_path


class _Args:
    def __init__(self, alignment_path, outdir, genbank_path, **overrides):
        self.alignment = str(alignment_path)
        self.output_dir = str(outdir)
        self.genbank = str(genbank_path)
        self.reference_id = "ref"
        self.n_threshold = 0.05
        self.cluster_window = 10
        self.cluster_count = 3
        self.mask_clustered = True
        self.mask_n_adjacent = True
        self.mask_gap_adjacent = True
        self.mask_frame_break = True
        for key, value in overrides.items():
            setattr(self, key, value)


def test_alignment_qc_end_to_end(tmp_path):
    aln_path, gb_path = _build_alignment(tmp_path)
    outdir = tmp_path / "out"

    args = _Args(aln_path, outdir, gb_path)
    result = alignment.main(args)
    assert result == 0

    summary_path = outdir / "alignment_qc_summary.txt"
    mask_path = outdir / "mask_sites.csv"
    assert summary_path.exists()
    assert mask_path.exists()

    rows = list(csv.DictReader(open(mask_path)))
    assert rows, "Expected mask_sites.csv to contain at least one row"

    site_rows = [row for row in rows if row.get("type") == "site"]
    notes_by_site = [row["note"] for row in site_rows]
    present_by_site = [row["present_in"] for row in site_rows]

    assert any("frame_break" in note for note in notes_by_site)
    assert any("s_frame" in present for present in present_by_site)

    assert any("clustered_snps" in note for note in notes_by_site)
    assert any("s_cluster" in present for present in present_by_site)


def test_alignment_qc_disable_n_adjacent(tmp_path):
    aln_path, gb_path = _build_alignment(tmp_path)
    outdir = tmp_path / "out"

    args = _Args(aln_path, outdir, gb_path, mask_n_adjacent=False)
    result = alignment.main(args)
    assert result == 0

    mask_path = outdir / "mask_sites.csv"
    rows = list(csv.DictReader(open(mask_path)))
    notes_by_site = [row["note"] for row in rows if row.get("type") == "site"]

    assert not any("N_adjacent" in note for note in notes_by_site)





import csv

from Bio import AlignIO
from Bio.Align import MultipleSeqAlignment
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

from raccoon.commands import mask


def _write_alignment(tmp_path):
    records = [
        SeqRecord(Seq("ACGTA"), id="s1"),
        SeqRecord(Seq("ACGTA"), id="s2"),
    ]
    aln = MultipleSeqAlignment(records)
    aln_path = tmp_path / "alignment.fasta"
    AlignIO.write(aln, str(aln_path), "fasta")
    return aln_path


def _write_mask(tmp_path):
    mask_path = tmp_path / "mask_sites.csv"
    with mask_path.open("w", newline="") as handle:
        writer = csv.DictWriter(handle, fieldnames=["flagged", "type", "Minimum", "Maximum", "Length", "present_in", "note"], lineterminator="\n")
        writer.writeheader()
        writer.writerow({
            "flagged": "2-3",
            "type": "site",
            "Minimum": "2",
            "Maximum": "3",
            "Length": "2",
            "present_in": "s1",
            "note": "masked",
        })
    return mask_path


class _Args:
    def __init__(self, alignment_path, mask_file, outdir, **overrides):
        self.alignment = str(alignment_path)
        self.mask_file = str(mask_file)
        self.outdir = str(outdir)
        self.output = None
        self.sequence_type = "nt"
        for key, value in overrides.items():
            setattr(self, key, value)


def test_mask_command_writes_masked_alignment(tmp_path):
    aln_path = _write_alignment(tmp_path)
    mask_path = _write_mask(tmp_path)
    outdir = tmp_path / "out"

    args = _Args(aln_path, mask_path, outdir)
    result = mask.main(args)
    assert result == 0

    masked_path = outdir / "alignment.masked.fasta"
    assert masked_path.exists()

    masked_aln = AlignIO.read(str(masked_path), "fasta")
    for rec in masked_aln:
        assert rec.seq[1] == "N"
        assert rec.seq[2] == "N"


def test_mask_command_preserves_unmasked_sites(tmp_path):
    records = [
        SeqRecord(Seq("ACGTAC"), id="s1"),
        SeqRecord(Seq("ACGTAC"), id="s2"),
    ]
    aln = MultipleSeqAlignment(records)
    aln_path = tmp_path / "alignment.fasta"
    AlignIO.write(aln, str(aln_path), "fasta")

    mask_path = tmp_path / "mask_sites.csv"
    with mask_path.open("w", newline="") as handle:
        writer = csv.DictWriter(handle, fieldnames=["flagged", "type", "Minimum", "Maximum", "Length", "present_in", "note"], lineterminator="\n")
        writer.writeheader()
        writer.writerow({
            "flagged": "5",
            "type": "site",
            "Minimum": "",
            "Maximum": "",
            "Length": "1",
            "present_in": "s2",
            "note": "masked",
        })

    outdir = tmp_path / "out"
    args = _Args(aln_path, mask_path, outdir)
    result = mask.main(args)
    assert result == 0

    masked_path = outdir / "alignment.masked.fasta"
    masked_aln = AlignIO.read(str(masked_path), "fasta")
    for rec in masked_aln:
        assert str(rec.seq)[:4] == "ACGT"
        assert rec.seq[4] == "N"
        assert rec.seq[5] == "C"


def test_mask_command_removes_sequence_records(tmp_path):
    records = [
        SeqRecord(Seq("ACGTAC"), id="s1"),
        SeqRecord(Seq("ACGTAC"), id="s2"),
    ]
    aln = MultipleSeqAlignment(records)
    aln_path = tmp_path / "alignment.fasta"
    AlignIO.write(aln, str(aln_path), "fasta")

    mask_path = tmp_path / "mask_sites.csv"
    with mask_path.open("w", newline="") as handle:
        writer = csv.DictWriter(handle, fieldnames=["flagged", "type", "Minimum", "Maximum", "Length", "present_in", "note"], lineterminator="\n")
        writer.writeheader()
        writer.writerow({
            "flagged": "s2",
            "type": "sequence_record",
            "Minimum": "1",
            "Maximum": "6",
            "Length": "6",
            "present_in": "s2",
            "note": "2,3,4",
        })

    outdir = tmp_path / "out"
    args = _Args(aln_path, mask_path, outdir)
    result = mask.main(args)
    assert result == 0

    masked_path = outdir / "alignment.masked.fasta"
    masked_aln = AlignIO.read(str(masked_path), "fasta")
    ids = [rec.id for rec in masked_aln]
    assert ids == ["s1"]



import csv
import textwrap
from pathlib import Path

from raccoon.commands import combine


class MockArgs:
    def __init__(self, **kwargs):
        self.inputs = kwargs.get("inputs", [])
        self.output = kwargs.get("output", None)
        self.metadata = kwargs.get("metadata", None)
        self.metadata_delimiter = kwargs.get("metadata_delimiter", ",")
        self.metadata_id_field = kwargs.get("metadata_id_field", "id")
        self.metadata_location_field = kwargs.get("metadata_location_field", "location")
        self.metadata_date_field = kwargs.get("metadata_date_field", "date")
        self.header_separator = kwargs.get("header_separator", "|")
        self.id_delimiter = kwargs.get("id_delimiter", "|")
        self.id_field = kwargs.get("id_field", 0)
        self.min_length = kwargs.get("min_length", None)
        self.max_n_content = kwargs.get("max_n_content", None)


def _write_fasta(path, entries):
    lines = []
    for header, seq in entries:
        lines.append(f">{header}")
        lines.append(seq)
    path.write_text("\n".join(lines) + "\n")


def test_combine_uppercase_and_unwrapped(tmp_path):
    f1 = tmp_path / "a.fasta"
    f2 = tmp_path / "b.fasta"

    _write_fasta(f1, [("seq1", "acgtacgt"), ("seq2", "aaaa")])
    _write_fasta(f2, [("seq3", "tttt")])

    out = tmp_path / "combined.fasta"
    args = MockArgs(inputs=[str(f1), str(f2)], output=str(out))

    result = combine.main(args)
    assert result == 0

    lines = out.read_text().strip().splitlines()
    assert lines[0] == ">seq1"
    assert lines[1] == "ACGTACGT"
    assert lines[2] == ">seq2"
    assert lines[3] == "AAAA"
    assert lines[4] == ">seq3"
    assert lines[5] == "TTTT"


def test_combine_harmonises_headers_with_metadata(tmp_path):
    fasta_path = tmp_path / "a.fasta"
    _write_fasta(fasta_path, [("seq1", "acgt"), ("seq2", "gggg")])

    metadata_path = tmp_path / "meta.csv"
    metadata_path.write_text(
        textwrap.dedent(
            """\
            id,location,date
            seq1,UK,2024-01-01
            seq2,US,2024-02-02
            """
        )
    )

    out = tmp_path / "combined.fasta"
    args = MockArgs(
        inputs=[str(fasta_path)],
        output=str(out),
        metadata=[str(metadata_path)],
    )

    result = combine.main(args)
    assert result == 0

    lines = out.read_text().strip().splitlines()
    assert lines[0] == ">seq1|UK|2024-01-01"
    assert lines[1] == "ACGT"
    assert lines[2] == ">seq2|US|2024-02-02"
    assert lines[3] == "GGGG"


def test_combine_examples_same_headers(tmp_path):
    repo_root = Path(__file__).resolve().parents[1]
    examples = repo_root / "tests" / "test_data" / "combine"
    out = tmp_path / "combined_same_headers.fasta"

    args = MockArgs(
        inputs=[
            str(examples / "inputs" / "set_a.fasta"),
            str(examples / "inputs" / "set_b.fasta"),
        ],
        output=str(out),
    )

    result = combine.main(args)
    assert result == 0

    lines = out.read_text().strip().splitlines()
    assert len(lines) == 16
    headers = lines[0::2]
    expected_headers = [
        ">A001",
        ">A002",
        ">A003",
        ">A004",
        ">B001",
        ">B002",
        ">B003",
        ">B004",
    ]
    assert headers == expected_headers
    for sequence in lines[1::2]:
        assert sequence == sequence.upper()
        assert " " not in sequence


def test_combine_examples_harmonised_headers(tmp_path):
    repo_root = Path(__file__).resolve().parents[1]
    examples = repo_root / "tests" / "test_data" / "combine"
    out = tmp_path / "combined_harmonised_headers.fasta"

    args = MockArgs(
        inputs=[
            str(examples / "inputs" / "set_a.fasta"),
            str(examples / "inputs" / "set_b.fasta"),
        ],
        output=str(out),
        metadata=[str(examples / "metadata.csv")],
    )

    result = combine.main(args)
    assert result == 0

    lines = out.read_text().strip().splitlines()
    assert len(lines) == 16
    headers = lines[0::2]
    expected_headers = [
        ">A001|SiteA|2024-01-15",
        ">A002|SiteA|2024-01-20",
        ">A003|SiteC|2024-02-01",
        ">A004|SiteC|2024-02-15",
        ">B001|LocX|2023-11-05",
        ">B002|LocX|2023-12-12",
        ">B003|LocZ|2024-03-03",
        ">B004|LocZ|2024-03-10",
    ]
    assert headers == expected_headers


def test_combine_multiple_metadata_files(tmp_path):
    fasta_path = tmp_path / "a.fasta"
    _write_fasta(fasta_path, [("seq1", "acgt"), ("seq2", "gggg")])

    meta1 = tmp_path / "meta1.csv"
    meta1.write_text(
        textwrap.dedent(
            """\
            id,location,date
            seq1,UK,2024-01-01
            """
        )
    )
    meta2 = tmp_path / "meta2.csv"
    meta2.write_text(
        textwrap.dedent(
            """\
            id,location,date
            seq2,US,2024-02-02
            """
        )
    )

    out = tmp_path / "combined.fasta"
    args = MockArgs(
        inputs=[str(fasta_path)],
        output=str(out),
        metadata=[str(meta1), str(meta2)],
    )

    result = combine.main(args)
    assert result == 0

    lines = out.read_text().strip().splitlines()
    assert lines[0] == ">seq1|UK|2024-01-01"
    assert lines[1] == "ACGT"
    assert lines[2] == ">seq2|US|2024-02-02"
    assert lines[3] == "GGGG"


def test_combine_parses_id_from_header(tmp_path):
    fasta_path = tmp_path / "a.fasta"
    _write_fasta(fasta_path, [("sample1|Loc1|2024-01-01", "acgt")])

    metadata_path = tmp_path / "meta.csv"
    metadata_path.write_text(
        textwrap.dedent(
            """\
            id,location,date
            sample1,UK,2024-01-01
            """
        )
    )

    out = tmp_path / "combined.fasta"
    args = MockArgs(
        inputs=[str(fasta_path)],
        output=str(out),
        metadata=[str(metadata_path)],
        id_delimiter="|",
        id_field=0,
    )

    result = combine.main(args)
    assert result == 0
    text = out.read_text()
    assert text.startswith(">sample1|UK|2024-01-01")


def test_combine_id_field_out_of_range_keeps_full_id(tmp_path):
    fasta_path = tmp_path / "a.fasta"
    _write_fasta(fasta_path, [("sample1|Loc1|2024-01-01", "acgt")])

    out = tmp_path / "combined.fasta"
    args = MockArgs(
        inputs=[str(fasta_path)],
        output=str(out),
        id_delimiter="|",
        id_field=10,
    )

    result = combine.main(args)
    assert result == 0
    text = out.read_text()
    assert text.startswith(">sample1|Loc1|2024-01-01")


def _read_csv_rows(path: Path):
    rows = []
    with path.open(newline="") as handle:
        reader = csv.DictReader(handle)
        for row in reader:
            rows.append(row)
    return rows


def test_combine_writes_filter_and_metadata_issue_csvs(tmp_path):
    fasta_path = tmp_path / "inputs.fasta"
    _write_fasta(
        fasta_path,
        [
            ("seq1", "AAAA"),
            ("seq2", "NNNN"),
            ("seq3", "ACGTACGT"),
        ],
    )

    metadata_path = tmp_path / "meta.tsv"
    metadata_path.write_text(
        textwrap.dedent(
            """\
            id\tlocation\tdate
            seq1\tUSA\t2024-01-01
            seq2\t\t2024-02-02
            """
        )
    )

    out = tmp_path / "combined.fasta"
    args = MockArgs(
        inputs=[str(fasta_path)],
        output=str(out),
        metadata=[str(metadata_path)],
        metadata_delimiter=",",
        min_length=5,
        max_n_content=0.2,
    )

    result = combine.main(args)
    assert result == 0

    filter_csv = tmp_path / "seq_qc_filter_failures.csv"
    metadata_csv = tmp_path / "seq_qc_metadata_issues.csv"
    assert filter_csv.exists()
    assert metadata_csv.exists()

    filter_rows = _read_csv_rows(filter_csv)
    filter_by_id = {row["id"]: row for row in filter_rows}
    assert set(filter_by_id) == {"seq1", "seq2"}
    assert filter_by_id["seq1"]["reason"] == "length < 5"
    assert filter_by_id["seq2"]["reason"] == "length < 5; N content > 0.2"

    metadata_rows = _read_csv_rows(metadata_csv)
    issues = {(row["id"], row["issue"], row["status"]) for row in metadata_rows}
    assert ("seq2", "missing location", "filtered") in issues
    assert ("seq3", "missing metadata row", "kept") in issues



import csv
import os
import types
import pytest
from pathlib import Path

from raccoon.utils import plotting


class FakeNode:
    def __init__(self, name, height, y, parent=None, branchType='leaf'):
        self.name = name
        self.height = height
        self.y = y
        self.parent = parent
        self.branchType = branchType
        self.traits = {}

    def is_leaf(self):
        return self.branchType == 'leaf'


class FakeTree:
    def __init__(self):
        # simple two-leaf tree
        root = FakeNode('root', height=0.0, y=0.0, parent=None, branchType='internal')
        a = FakeNode('A', height=1.0, y=1.0, parent=root, branchType='leaf')
        b = FakeNode('B', height=1.0, y=2.0, parent=root, branchType='leaf')
        self.Objects = [a, b, root]
        self.treeHeight = 1.0
        self.ySpan = 10.0

    def plotTree(self, ax, x_attr=None):
        # no-op for tests
        pass

    def plotPoints(self, ax, size=None, colour=None, x_attr=None):
        pass

    def addText(self, ax=None, x_attr=None, target=None, text=None):
        pass


class FakeBalticModule:
    def loadNewick(self, treefile, absoluteTime=False):
        return FakeTree()

    def loadNexus(self, treefile, absoluteTime=False):
        return FakeTree()


@pytest.fixture(autouse=True)
def patch_baltic(monkeypatch):
    """Monkeypatch plotting.bt to a fake baltic module for tests."""
    fake = FakeBalticModule()
    monkeypatch.setattr(plotting, 'bt', fake)
    yield


def write_branch_snps_csv(path):
    rows = [
        {'parent': 'root', 'child': 'A', 'site': '10', 'snp': 'G->A', 'dimer': 'GA'},
        {'parent': 'root', 'child': 'B', 'site': '20', 'snp': 'C->T', 'dimer': 'TC'},
    ]
    with open(path, 'w') as f:
        writer = csv.DictWriter(f, fieldnames=['parent', 'child', 'site', 'snp', 'dimer'])
        writer.writeheader()
        for r in rows:
            writer.writerow(r)


def test_make_reversion_tree_figure_writes_files(tmp_path):
    base = tmp_path / 'rev_out'
    csv_path = tmp_path / 'branch_snps.csv'
    write_branch_snps_csv(csv_path)

    branch_reversions = {'root_A': {'10A'}}
    will_be_reverted = {'root_A': {'5G'}}

    plotting.make_reversion_tree_figure(str(base), str(csv_path), branch_reversions, will_be_reverted, treefile='tree', w=None, h=None)

    assert base.with_suffix('.svg').exists()
    assert base.with_suffix('.png').exists()


def test_make_reversion_tree_figure_accepts_dict(tmp_path):
    base = tmp_path / 'rev_out2'
    branch_snps_dict = {
        'root_A': [('10', 'G->A', 'GA')],
        'root_B': [('20', 'C->T', 'TC')],
    }
    plotting.make_reversion_tree_figure(str(base), branch_snps_dict, {}, {}, treefile='tree', w=10, h=4)
    assert base.with_suffix('.svg').exists()
    assert base.with_suffix('.png').exists()


def test_make_convergence_tree_figure(tmp_path):
    base = tmp_path / 'conv_out'
    branch_snps_dict = {
        'root_A': [('10', 'G->A', 'GA')],
    }
    branch_convergence = {'root_A': {'G10A'}}
    plotting.make_convergence_tree_figure(str(base), branch_snps_dict, branch_convergence, treefile='tree', w=None, h=None)
    assert base.with_suffix('.svg').exists()
    assert base.with_suffix('.png').exists()


def test_make_reconstruction_tree_figure_w_labels(tmp_path):
    base = tmp_path / 'rec_out'
    branch_snps_dict = {
        'root_A': [('10', 'G->A', 'GA')],
    }
    plotting.make_reconstruction_tree_figure_w_labels(str(base), branch_snps_dict, treefile='tree', point_style='circle', justification='right', w=10, h=5)
    assert base.with_suffix('.svg').exists()
    assert base.with_suffix('.png').exists()


def test_functions_require_baltic(monkeypatch):
    # ensure functions raise when baltic not available
    monkeypatch.setattr(plotting, 'bt', None)
    with pytest.raises(RuntimeError):
        plotting.make_convergence_tree_figure('x', {}, {}, 'tree', None, None)
    with pytest.raises(RuntimeError):
        plotting.make_reversion_tree_figure('x', {}, {}, {}, 'tree', None, None)
    with pytest.raises(RuntimeError):
        plotting.make_reconstruction_tree_figure_w_labels('x', {}, 'tree', 'circle', 'left', None, None)



#!/usr/bin/env python3
END_FORMATTING = '\033[0m'
RED = '\033[31m'
GREEN = '\033[32m'
YELLOW = '\033[93m'
CYAN = '\u001b[36m'
DIM = '\033[2m'

def red(text):
    return RED + text + END_FORMATTING

def cyan(text):
    return CYAN + text + END_FORMATTING

def green(text):
    return GREEN + text + END_FORMATTING

def yellow(text):
    return YELLOW + text + END_FORMATTING



"""Tests for raccoon.utils.io module."""
import os
import tempfile
import pytest
from raccoon.utils import io


class TestEnsureOutputDirectory:
    """Tests for ensure_output_directory function."""

    def test_creates_nonexistent_directory(self, tmp_path):
        """Test that function creates a directory if it doesn't exist."""
        new_dir = tmp_path / "new_output"
        assert not new_dir.exists()
        
        result = io.ensure_output_directory(str(new_dir))
        
        assert result is True
        assert new_dir.exists()
        assert new_dir.is_dir()

    def test_accepts_existing_directory(self, tmp_path):
        """Test that function accepts an existing writable directory."""
        result = io.ensure_output_directory(str(tmp_path))
        
        assert result is True

    def test_rejects_readonly_directory(self, tmp_path):
        """Test that function rejects a read-only directory."""
        readonly_dir = tmp_path / "readonly"
        readonly_dir.mkdir()
        os.chmod(str(readonly_dir), 0o444)
        
        try:
            result = io.ensure_output_directory(str(readonly_dir))
            assert result is False
        finally:
            # restore permissions for cleanup
            os.chmod(str(readonly_dir), 0o755)


class TestValidateInputFile:
    """Tests for validate_input_file function."""

    def test_accepts_existing_readable_file(self, tmp_path):
        """Test that function accepts an existing readable file."""
        test_file = tmp_path / "test.txt"
        test_file.write_text("content")
        
        result = io.validate_input_file(str(test_file))
        
        assert result is True

    def test_rejects_nonexistent_file(self, tmp_path):
        """Test that function rejects a nonexistent file."""
        result = io.validate_input_file(str(tmp_path / "nonexistent.txt"))
        
        assert result is False

    def test_accepts_empty_filepath(self):
        """Test that function accepts empty/None filepath for optional files."""
        result = io.validate_input_file("")
        
        assert result is True

    def test_rejects_readonly_file(self, tmp_path):
        """Test that function rejects a read-only file."""
        test_file = tmp_path / "readonly.txt"
        test_file.write_text("content")
        os.chmod(str(test_file), 0o000)
        
        try:
            result = io.validate_input_file(str(test_file))
            assert result is False
        finally:
            # restore permissions for cleanup
            os.chmod(str(test_file), 0o644)


class TestValidateAlignmentFile:
    """Tests for validate_alignment_file function."""

    def test_accepts_existing_alignment_file(self, tmp_path):
        """Test that function accepts an existing alignment file."""
        aln_file = tmp_path / "alignment.fasta"
        aln_file.write_text(">seq1\nACGT")
        
        result = io.validate_alignment_file(str(aln_file))
        
        assert result is True

    def test_rejects_nonexistent_alignment_file(self, tmp_path):
        """Test that function rejects a nonexistent alignment file."""
        result = io.validate_alignment_file(str(tmp_path / "nonexistent.fasta"))
        
        assert result is False


class TestValidateGenbankFile:
    """Tests for validate_genbank_file function."""

    def test_accepts_existing_genbank_file(self, tmp_path):
        """Test that function accepts an existing GenBank file."""
        gb_file = tmp_path / "reference.gb"
        gb_file.write_text("LOCUS test")
        
        result = io.validate_genbank_file(str(gb_file))
        
        assert result is True

    def test_accepts_empty_filepath_optional(self):
        """Test that function accepts empty filepath (optional)."""
        result = io.validate_genbank_file("")
        
        assert result is True

    def test_rejects_nonexistent_genbank_file(self, tmp_path):
        """Test that function rejects a nonexistent GenBank file."""
        result = io.validate_genbank_file(str(tmp_path / "nonexistent.gb"))
        
        assert result is False


class TestValidateReferenceFile:
    """Tests for validate_reference_file function."""

    def test_accepts_existing_reference_file(self, tmp_path):
        """Test that function accepts an existing reference file."""
        ref_file = tmp_path / "reference.fasta"
        ref_file.write_text(">ref\nACGT")
        
        result = io.validate_reference_file(str(ref_file))
        
        assert result is True

    def test_accepts_empty_filepath_optional(self):
        """Test that function accepts empty filepath (optional)."""
        result = io.validate_reference_file("")
        
        assert result is True

    def test_rejects_nonexistent_reference_file(self, tmp_path):
        """Test that function rejects a nonexistent reference file."""
        result = io.validate_reference_file(str(tmp_path / "nonexistent.fasta"))
        
        assert result is False



#!/usr/bin/env python3
"""Mask subcommand wrapper."""
import os
import logging


def main(args):
    if not hasattr(args, "alignment"):
        raise ValueError("Expected argparse Namespace from raccoon.command.build_parser")

    logging.info("Starting mask")
    try:
        from raccoon.utils import alignment_functions as af
        from raccoon.utils import io

        outdir = args.outdir or os.getcwd()
        if not io.ensure_output_directory(outdir):
            return 1

        alignment_path, temp_path = io.prepare_alignment_input(args.alignment)
        try:
            if not io.validate_alignment_file(alignment_path):
                return 1

            mask_file = getattr(args, "mask_file", None)
            if not io.validate_input_file(mask_file, "Mask file"):
                return 1

            output_path = args.output
            if not output_path:
                base = os.path.splitext(os.path.basename(alignment_path))[0]
                output_path = f"{base}.masked.fasta"

            if not os.path.isabs(output_path):
                output_path = os.path.join(outdir, output_path)

            mask_char = "X" if getattr(args, "sequence_type", "nt") == "aa" else "N"
            masked_sites = af.apply_mask_to_alignment(
                alignment_path,
                mask_file,
                output_path,
                mask_char=mask_char,
            )
            logging.info(f"Masked alignment written to {output_path} ({masked_sites} sites masked)")
            try:
                from raccoon.utils import reporting
                reporting.generate_mask_report(
                    outdir=outdir,
                    alignment_path=alignment_path,
                    mask_file=mask_file,
                    output_alignment=output_path,
                )
            except Exception:
                logging.exception("Failed to generate mask report")
            logging.info("Mask finished")
            return 0
        finally:
            io.cleanup_temp_file(temp_path)
    except FileNotFoundError as exc:
        logging.error(str(exc))
        return 1
    except Exception:
        logging.exception("Mask failed")
        return 2



import pytest

from raccoon.command import build_parser


def test_subcommands_present():
    parser = build_parser()
    # access to subparser choices (stable across argparse implementations)
    subparsers = parser._subparsers._group_actions[0].choices
    assert 'aln-qc' in subparsers
    assert 'tree-qc' in subparsers
    assert 'seq-qc' in subparsers


def test_alignment_sets_func_callable():
    parser = build_parser()
    ns = parser.parse_args(['aln-qc', 'input.fasta'])
    assert callable(ns.func)


def test_phylo_sets_func_callable():
    parser = build_parser()
    ns = parser.parse_args(['tree-qc', '--phylogeny', 'tree'])
    assert callable(ns.func)


def test_combine_sets_func_callable():
    parser = build_parser()
    ns = parser.parse_args(['seq-qc', 'a.fasta'])
    assert callable(ns.func)



import textwrap
import os

from Bio import AlignIO
from Bio.Align import MultipleSeqAlignment
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio.SeqFeature import SeqFeature, FeatureLocation

from raccoon.utils import alignment_functions as af


def test_find_high_n_sequences(tmp_path):
    r1 = SeqRecord(Seq('ATGATGATGATG'), id='seq1')
    r2 = SeqRecord(Seq('ATGNNNNNNATG'), id='seq2')
    aln = MultipleSeqAlignment([r1, r2])
    path = tmp_path / 'test.fasta'
    AlignIO.write(aln, str(path), 'fasta')

    aln2 = AlignIO.read(str(path), 'fasta')
    flagged = af.find_high_N_sequences(aln2, threshold=0.3)
    assert len(flagged) == 1
    assert flagged[0][0] == 'seq2'


def test_analyze_alignment_unique_snps_and_flags():
    # alignment where s2 has three unique SNPs, one near N and one near a gap
    ref_seq = 'AAAAAAAAAAAA'
    ref = SeqRecord(Seq(ref_seq), id='ref')
    s1 = SeqRecord(Seq(ref_seq), id='s1')
    s2 = SeqRecord(Seq('AATAGNA-CAAA'), id='s2')

    aln = MultipleSeqAlignment([ref, s1, s2])
    unique_mutations, snps_near_n, snps_near_gap, clustered_snps = af.analyze_alignment(
        aln, n_window=2, gap_window=1, snp_window=10, snp_count=3
    )

    assert unique_mutations['s2'] == {2, 4, 8}
    assert snps_near_n['s2'] == {4}
    assert snps_near_gap['s2'] == {8}
    assert clustered_snps['s2'] == {2, 4, 8}


def test_sliding_window():
    assert list(af.sliding_window([1, 2, 3, 4], 2)) == [[1, 2], [2, 3], [3, 4]]
    assert list(af.sliding_window([1, 2], 3)) == [[1, 2]]


def test_find_frame_breaking_indels(tmp_path):
    # reference sequence with a CDS covering the full length
    ref_record = SeqRecord(Seq('ATGAAATTT'), id='ref', name='ref')
    ref_record.annotations['molecule_type'] = 'DNA'
    ref_record.features.append(SeqFeature(FeatureLocation(0, 9), type='CDS'))

    genbank_path = tmp_path / 'ref.gb'
    with open(genbank_path, 'w') as handle:
        from Bio import SeqIO
        SeqIO.write(ref_record, handle, 'genbank')

    # alignment where seq2 has a single gap in the CDS region -> frame break
    ref_aln = SeqRecord(Seq('ATGAAATTT'), id='ref')
    seq2 = SeqRecord(Seq('ATG-AATTT'), id='seq2')
    aln = MultipleSeqAlignment([ref_aln, seq2])

    sites = af.find_frame_breaking_indels(aln, str(genbank_path), reference_id='ref')
    assert sites, "Expected frame-breaking sites to be reported"
    # ensure at least one site includes seq2 in present_in
    assert any('seq2' in v['present_in'] for v in sites.values())



#!/usr/bin/env python3
"""Alignment subcommand wrapper."""
import os
import logging

from raccoon.utils import constants as rc

def main(args):
    """Run alignment QC using functions from raccoon.utils.

    Expects an argparse Namespace with attributes added by the top-level parser.
    """
    if not hasattr(args, 'alignment'):
        raise ValueError("Expected argparse Namespace from raccoon.command.build_parser")

    logging.info("Starting alignment QC")
    try:
        # lazy import to keep module lightweight at import-time
        from raccoon.utils import alignment_functions as af
        from raccoon.utils import io

        outdir = args.output_dir or os.getcwd()
        
        # validate output directory
        if not io.ensure_output_directory(outdir):
            return 1
        
        # validate input files
        if not io.validate_alignment_file(args.alignment):
            return 1
        
        genbank = getattr(args, rc.KEY_GENBANK, None)
        if genbank and not io.validate_genbank_file(genbank):
            return 1
        
        reference = getattr(args, rc.KEY_REFERENCE_ID, None)
        n_threshold = getattr(args, rc.KEY_N_THRESHOLD, rc.DEFAULT_N_THRESHOLD)
        cluster_window = getattr(args, rc.KEY_CLUSTER_WINDOW, rc.DEFAULT_CLUSTER_WINDOW)
        cluster_count = getattr(args, rc.KEY_CLUSTER_COUNT, rc.DEFAULT_CLUSTER_COUNT)
        mask_clustered = getattr(args, rc.KEY_MASK_CLUSTERED, True) 
        mask_n_adjacent = getattr(args, rc.KEY_MASK_N_ADJACENT, True) 
        mask_gap_adjacent = getattr(args, rc.KEY_MASK_GAP_ADJACENT, True) 
        mask_frame_break = getattr(args, rc.KEY_MASK_FRAME_BREAK, True)

        summary = af.run_alignment_qc(
            args.alignment,
            outdir=outdir,
            genbank_path=genbank,
            reference_id=reference,
            n_threshold=n_threshold,
            cluster_window=cluster_window,
            cluster_count=cluster_count,
            mask_clustered=mask_clustered,
            mask_n_adjacent=mask_n_adjacent,
            mask_gap_adjacent=mask_gap_adjacent,
            mask_frame_break=mask_frame_break,
        )


        logging.info("Alignment QC completed")
        logging.info(f"High N sequences: {len(summary['high_n_sequences'])}")
        logging.info(f"Sites to mask: {len(summary['sites_to_mask'])}")

        # write a summary file
        summary_file = os.path.join(outdir, 'alignment_qc_summary.txt')
        with open(summary_file, 'w') as fw:
            fw.write(f"issues_found: {summary['issues_found']}\n")
            fw.write(f"high_n_sequences: {len(summary['high_n_sequences'])}\n")
            fw.write(f"sites_to_mask: {len(summary['sites_to_mask'])}\n")
            fw.write(f"mask_file: {summary['mask_file']}\n")

        logging.info(f"Summary written to {summary_file}")

        try:
            from raccoon.utils import reporting
            reporting.generate_alignment_report(
                outdir=outdir,
                alignment_path=args.alignment,
                mask_file=summary.get(rc.KEY_MASK_FILE),
            )
        except Exception:
            logging.exception("Failed to generate alignment report")

        return 0
    except FileNotFoundError as exc:
        logging.error(str(exc))
        return 1
    except Exception:
        logging.exception("Alignment QC failed")
        return 2



#!/usr/bin/env python3
import sys
import os
import logging
try:
    from raccoon.utils.misc import green, cyan
except Exception:
    def green(s):
        return s
    def cyan(s):
        return s
import statistics
import select
from Bio import SeqIO
from Bio import AlignIO
from Bio.Align.AlignInfo import SummaryInfo
import collections
import csv
from .constants import KEY_ASSEMBLY_REFERENCES, KEY_OUTDIR, KEY_OUTFILENAME, KEY_PHYLOGENY, KEY_RUN_APOBEC3_PHYLO
import math
import matplotlib as mpl
import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd

from raccoon.utils.reconstruction_functions import load_tree, ensure_node_label


def find_assembly_refs(cwd,assembly_refs,config):
    refs = []
    ref_ids = []
    if not assembly_refs:
        logging.info(cyan(f'Note: no assembly references supplied.\nDefaulting to installed assembly references:'))
        for record in SeqIO.parse(config[KEY_ASSEMBLY_REFERENCES],"fasta"):
            refs.append(record)
            ref_ids.append(record.id)
        for i in ref_ids:
            logging.info(f"- {i}")
    else:
        path_to_try = os.path.join(cwd,assembly_refs)
        try:
            for record in SeqIO.parse(path_to_try,"fasta"):
                refs.append(record)
                ref_ids.append(record.id)
                
        except Exception as exc:
            msg = cyan(f'Error: cannot find/parse reference fasta file at: ') + f'{path_to_try}\n' + cyan('Please check file path and format.\n')
            logging.error(msg)
            raise FileNotFoundError(msg) from exc

        logging.info(green(f'Assembly references supplied:'))
        for i in ref_ids:
            logging.info(f"- {i}")

    config[KEY_ASSEMBLY_REFERENCES] = ref_ids

    return refs

def recurse_back(node,root_node):
    path = [node.traits["label"]]
    p = node.parent
    root_name = root_node.traits["label"]
    while p.traits["label"] != root_name:
        path.append(p.traits["label"])
        p = p.parent

    path.append(root_name)
    return path

def get_path_to_root(treefile, tree_format="auto"):
    
    """
    traces back from a given tip through each parent 
    and returns a list of branches that make up the 
    phylo path from root to tip

    """
    my_tree = load_tree(treefile, tree_format=tree_format)
    root_node = ""
    for k in my_tree.Objects:
        current_node = k
        if not root_node:
            root_node = k
            ensure_node_label(root_node)

        ensure_node_label(current_node)
        if getattr(current_node, "parent", None):
            ensure_node_label(current_node.parent)
            
    
    branch_paths = {}
    
    for k in my_tree.Objects:
        current_node = k
        if current_node.branchType == 'leaf':
            ensure_node_label(current_node)
            path = recurse_back(current_node,root_node)
            branch_path = []
            for i in range(len(path)-1, 0, -1):
                try:
                    branch = f"{path[i]}_{path[i-1]}"
                    branch_path.append(branch)
                except Exception:
                    logging.debug(f"breaks {path}")
#             print(current_node.traits["label"])
#             print(branch_path)
            branch_paths[current_node.traits["label"]] = branch_path
    return branch_paths


def read_in_branch_snps(branch_snps):
    branch_snps_dict = collections.defaultdict(list)
    with open(branch_snps,"r") as f:
        reader = csv.DictReader(f)
        for row in reader:
            branch = f"{row['parent']}_{row['child']}"
            branch_snps_dict[branch].append((row['site'],row['snp'],row['dimer'])) 
    return branch_snps_dict

def get_seq_at_node(state_file,nodename):
    
    """
    returns a dict keys off 1-based positions in the genome
    and the value is a list of tuples (id, base) at a given node
    allows you to look up for a given site what the base is for a
    given internal node or tip
    """
    seq = ""
    with open(f"{state_file}","r") as f:
        for l in f:

            if not l.startswith("#"):
                try:
                    node,site,state,probA,probC,probG,probT = l.rstrip("\n").split("\t")
                except Exception:
                    logging.debug(l)
                    break
                if node == nodename:
                    seq+=state

    return seq

def load_assembly_refs(assembly_refs):

    refs = {}
    if isinstance(assembly_refs, str):
        for record in SeqIO.parse(assembly_refs, "fasta"):
            refs[record.id] = record.seq
        return refs
    for record in assembly_refs:
        refs[record.id] = record.seq
    return refs


def flag_reversions(branch_paths, branch_snp_dict,state_file, refs):
    root_node = get_seq_at_node(state_file,"Node1")
    possible_reversions = []
    branch_reversions = collections.defaultdict(set)
    snp_to_branch = {}
    will_be_reverted = collections.defaultdict(set)
    for tip in branch_paths:
        path_snps = []
        for branch in branch_paths[tip]:
            snps = branch_snp_dict[branch]
            for i in snps:
                
                if i[0] in [j[0] for j in path_snps]:
                    base = int(i[0])
                    
                    allele = i[1][-1]
                    
                    branch_reversions[branch].add(f"{base}{allele}")
                    will_be_reverted[snp_to_branch[i[0]]].add(f"{base}{i[1][0]}")
                    reversion_to = []
                    
                    ref_alleles = []
                    for ref in refs:
                        var = refs[ref][base-1]
                        ref_alleles.append(f"{ref}:{var}")
                        if allele == var:
                            reversion_to.append(ref)
                        
                    root_var = root_node[base-1]
                    if allele == root_var:
                        reversion_to.append("Root")
                    
                    original_snp = [k for k in path_snps if k[0] == i[0]][0]
                    
                    possible_reversions.append({
                        "taxon":tip,
                        "site":i[0],
                        "original_snp": original_snp[1],
                        "original_branch":snp_to_branch[i[0]],
                        "reversion_branch":branch,
                        "dinucleotide_context": original_snp[2],
                        "reversion_snp": i[1],
                        "reference_alleles": ";".join(ref_alleles),
                        "root_allele":root_node[base-1],
                        "reversion_to":";".join(reversion_to)
                    })

                
                snp_to_branch[i[0]] = branch
                path_snps.append(i)
    if branch_reversions:
        logging.info(green("Reversions flagged:"))
        for i in branch_reversions:
            for j in branch_reversions[i]:
                logging.info(f"- {j} ({i})")
    return possible_reversions,branch_reversions,will_be_reverted


def flag_convergence(treefile, branch_snp_dict, tree_format="auto"):
    snp_to_branch = collections.defaultdict(set)
    branch_convergence = collections.defaultdict(set)

    my_tree = load_tree(treefile, tree_format=tree_format)

    for k in my_tree.Objects:
        current_node = k
        ensure_node_label(current_node)
        if getattr(current_node, "parent", None):
            ensure_node_label(current_node.parent)
        if k.branchType == 'leaf':
            node_name = current_node.traits.get("label")
            
        try:
            parent_name = current_node.parent.traits.get("label")
            node_name = current_node.traits.get("label")
        except Exception:
            continue
        if not parent_name or not node_name:
            continue
        
        branch = f"{parent_name}_{node_name}"

        snps = branch_snp_dict[branch]
        for i in snps:
            snp_to_branch[i].add(branch)
    convergent_snps= []
    for snp in snp_to_branch:
        if len(snp_to_branch[snp])> 1:
            report_snp = f"{snp[1][0]}{snp[0]}{snp[1][-1]}"
            convergent_snps.append(report_snp)
            for branch in snp_to_branch[snp]:
                branch_convergence[branch].add(report_snp)

    if convergent_snps:
        logging.info("Convergent snps flagged:")
        for i in convergent_snps:
            logging.info(f"- {i}")
    return branch_convergence


def make_reversion_tree_figure(outfile,branch_snps,branch_reversions,will_be_reverted,treefile,w,h):
    """Wrapper that delegates to plotting.make_reversion_tree_figure"""
    from .plotting import make_reversion_tree_figure as _plot
    return _plot(outfile, branch_snps, branch_reversions, will_be_reverted, treefile, w, h)
    # plt.show()

def make_convergence_tree_figure(outfile,branch_snps,branch_convergence,treefile,w,h):
    """Wrapper that delegates to plotting.make_convergence_tree_figure"""
    from .plotting import make_convergence_tree_figure as _plot
    return _plot(outfile, branch_snps, branch_convergence, treefile, w, h)


def run_phylo_snp_checks(assembly_references,config,h):

    state_file = os.path.join(config[KEY_OUTDIR],f"{config[KEY_PHYLOGENY]}.state")
    treefile = os.path.join(config[KEY_OUTDIR],f"{config[KEY_PHYLOGENY]}")

    branch_snps = os.path.join(config[KEY_OUTDIR],f"{config[KEY_PHYLOGENY]}.branch_snps.reconstruction.csv")
    reversion_figure_out = os.path.join(config[KEY_OUTDIR],f"{config[KEY_OUTFILENAME]}.reversions_fig")
    convergence_figure_out = os.path.join(config[KEY_OUTDIR],f"{config[KEY_OUTFILENAME]}.convergence_fig")

    refs = load_assembly_refs(assembly_references)
    

    branch_snp_dict = read_in_branch_snps(branch_snps)
    branch_paths= get_path_to_root(treefile)

    possible_reversions,branch_reversions,will_be_reverted = flag_reversions(branch_paths, branch_snp_dict,state_file, refs)  
    branch_convergence = flag_convergence(treefile, branch_snp_dict)
    make_reversion_tree_figure(reversion_figure_out,branch_snps,branch_reversions,will_be_reverted,treefile,25,h)
    make_convergence_tree_figure(convergence_figure_out,branch_snps,branch_convergence,treefile,25,h)

    return branch_reversions, branch_convergence

def check_for_snp_anomalies(assembly_references,outgroup_ids,mask_file,config,h):

    alignment = os.path.join(config[KEY_OUTDIR],config[KEY_OUTFILENAME])
    branch_reversions, branch_convergence = {},{}

    if config[KEY_RUN_APOBEC3_PHYLO]:
        branch_reversions, branch_convergence = run_phylo_snp_checks(assembly_references,config,h)

    sites_to_mask = check_for_alignment_issues(alignment,outgroup_ids)

    merge_flagged_sites(sites_to_mask,branch_reversions,branch_convergence,mask_file)

def flag_long_branches(treefile, tree_format="auto", sd_threshold=3.0):
    my_tree = load_tree(treefile, tree_format=tree_format)
    tip_heights = []
    tip_names = []
    for node in my_tree.Objects:
        if node.branchType == 'leaf':
            ensure_node_label(node)
            tip_names.append(node.traits.get("label"))
            tip_heights.append(node.height)

    if not tip_heights:
        return {}

    mean_val = statistics.mean(tip_heights)
    std_val = statistics.pstdev(tip_heights)
    if std_val == 0:
        return {}

    flags = {}
    for name, height in zip(tip_names, tip_heights):
        if height > mean_val + (sd_threshold * std_val):
            flags[name] = height
    return flags

def classify_snp_for_edit(snp, dimer):
    if snp == "G->A" and dimer == "GA":
        return "apobec3"
    if snp == "C->T" and dimer == "TC":
        return "apobec3"
    if snp == "A->G" or snp == "T->C":
        return "adar"
    return None

def _has_adr_cluster(sites, window, min_count):
    if len(sites) < min_count:
        return False
    sites = sorted(sites)
    for i in range(len(sites) - min_count + 1):
        if sites[i + min_count - 1] - sites[i] <= window:
            return True
    return False

def parse_site_from_report(report_snp):
    digits = "".join([c for c in report_snp if c.isdigit()])
    return digits or ""

def build_branch_snps(treefile, state_file, alignment, outdir, tree_format, phylogeny_base):
    from raccoon.utils import reconstruction_functions as rf

    state_differences = os.path.join(outdir, f"{phylogeny_base}.state_differences.csv")
    branch_snps_out = os.path.join(outdir, f"{phylogeny_base}.branch_snps.reconstruction.csv")

    node_states = rf.generate_reconstruction_files(alignment, state_file, state_differences)
    node_states_diff = rf.load_unambiguous_varying_sites(state_differences)
    rf.map_site_changes_to_branches(treefile, branch_snps_out, node_states, node_states_diff, tree_format=tree_format)

    return branch_snps_out

def write_phylo_flags_csv(outfile, rows):
    def _merge_rows(items):
        merged = {}
        order = []
        for row in items:
            site = row.get("site", "")
            mutation_type = row.get("mutation_type", "")
            key = (site, mutation_type)
            if key not in merged:
                merged[key] = {
                    "site": site,
                    "mutation_type": mutation_type,
                    "present_in": set(),
                    "mask_boolean": False,
                }
                order.append(key)
            present = row.get("present_in", "")
            if present:
                merged[key]["present_in"].add(str(present))
            merged[key]["mask_boolean"] = merged[key]["mask_boolean"] or bool(row.get("mask_boolean", False))
        output = []
        for key in order:
            entry = merged[key]
            present = ";".join(sorted(entry["present_in"]))
            output.append({
                "site": entry["site"],
                "mutation_type": entry["mutation_type"],
                "present_in": present,
                "mask_boolean": entry["mask_boolean"],
            })
        return output

    merged_rows = _merge_rows(rows)
    with open(outfile, "w") as handle:
        writer = csv.DictWriter(handle, fieldnames=["site", "mutation_type", "present_in", "mask_boolean"], lineterminator="\n")
        writer.writeheader()
        for row in merged_rows:
            writer.writerow(row)

def run_phylo_qc(
    treefile,
    tree_format,
    outdir,
    alignment=None,
    state_file=None,
    assembly_refs=None,
    long_branch_sd=3.0,
    include_apobec=False,
    include_adar=False,
    adar_window=300,
    adar_min_count=3,
    output_csv=None,
):
    rows = []

    long_branch_flags = flag_long_branches(treefile, tree_format=tree_format, sd_threshold=long_branch_sd)
    for tip_name in sorted(long_branch_flags):
        rows.append({
            "site": "",
            "mutation_type": "long_branch",
            "present_in": tip_name,
            "mask_boolean": True,
        })

    if state_file:
        if not alignment:
            raise ValueError("Alignment is required when using an ASR state file")

        phylogeny_base = os.path.splitext(os.path.basename(treefile))[0]
        branch_snps = build_branch_snps(treefile, state_file, alignment, outdir, tree_format, phylogeny_base)
        branch_snp_dict = read_in_branch_snps(branch_snps)

        if include_apobec or include_adar:
            adar_sites_by_branch = collections.defaultdict(list)
            adar_all_by_branch = collections.defaultdict(list)
            for branch, snps in branch_snp_dict.items():
                for site, snp, dimer in snps:
                    edit_type = classify_snp_for_edit(snp, dimer)
                    if edit_type == "apobec3" and include_apobec:
                        rows.append({
                            "site": site,
                            "mutation_type": "apobec3",
                            "present_in": branch,
                            "mask_boolean": False,
                        })
                    elif edit_type == "adar" and include_adar:
                        try:
                            adar_sites_by_branch[branch].append(int(site))
                        except Exception:
                            continue
                        adar_all_by_branch[branch].append(site)

            if include_adar:
                adar_first_allowed = False
                for branch, sites in adar_sites_by_branch.items():
                    if _has_adr_cluster(sites, adar_window, adar_min_count):
                        ordered_sites = sorted(adar_all_by_branch[branch], key=lambda s: int(s))
                        for site in ordered_sites:
                            mask_value = True
                            if not adar_first_allowed:
                                mask_value = False
                                adar_first_allowed = True
                            rows.append({
                                "site": site,
                                "mutation_type": "adar",
                                "present_in": branch,
                                "mask_boolean": mask_value,
                            })

        branch_paths = get_path_to_root(treefile, tree_format=tree_format)
        refs = load_assembly_refs(assembly_refs) if assembly_refs else {}
        possible_reversions, branch_reversions, will_be_reverted = flag_reversions(
            branch_paths, branch_snp_dict, state_file, refs
        )

        for item in possible_reversions:
            rows.append({
                "site": item.get("site", ""),
                "mutation_type": "reversion",
                "present_in": item.get("taxon", ""),
                "mask_boolean": True,
            })

        branch_convergence = flag_convergence(treefile, branch_snp_dict, tree_format=tree_format)
        for branch, snps in branch_convergence.items():
            for report_snp in snps:
                rows.append({
                    "site": parse_site_from_report(report_snp),
                    "mutation_type": "convergent_snp",
                    "present_in": branch,
                    "mask_boolean": True,
                })

    output_csv = output_csv or os.path.join(outdir, f"{os.path.splitext(os.path.basename(treefile))[0]}.phylo_flags.csv")
    write_phylo_flags_csv(output_csv, rows)
    return output_csv




import os
import logging
try:
    from raccoon.utils.misc import green, cyan
except Exception:
    def green(s):
        return s
    def cyan(s):
        return s
import warnings
from Bio import BiopythonWarning
warnings.simplefilter('ignore', BiopythonWarning)

from Bio import AlignIO
from Bio import SeqIO
from Bio.Seq import Seq
from datetime import date
import datetime as dt

import csv

import baltic as bt
from scipy import stats

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
import collections
import pandas as pd
import matplotlib.patches as patches

import math
plt.switch_backend('Agg') 


mpl.rcParams.update({'font.size': 18})

new_rc_params = {'text.usetex': False,
"svg.fonttype": 'none'
}
mpl.rcParams.update(new_rc_params)
plt.rcParams['font.family'] = 'Helvetica'


def get_node_states_all_sites(state_file,alignment):
    
    #returns a dict keys off 1-based positions in the genome
    #and the value is a list of tuples (id, base) at a given node
    # allows you to look up for a given site what the base is for a
    #given internal node or tip
    
    node_states = collections.defaultdict(list)
    c = 0
    
    ## first the reconstructed nodes
    with open(state_file,"r") as f:
        for l in f:

            if not l.startswith("#"):
                c+=1
                try:
                    node,site,state,probA,probC,probG,probT = l.rstrip("\n").split("\t")
                except Exception:
                    logging.debug(l)
                    break
                if node != "Node":
                    if state not in ["N","-"]:
                        node_states[site].append((node,state))
                    else:
                        node_states[site].append((node,""))
    ## now the tips
    for record in SeqIO.parse(alignment,"fasta"):
        for site in node_states:
            index = int(site)-1
            base = record.seq[index].upper()
            if base in ["T","C","A","G"]:
                node_states[site].append((record.id,base))
            else:
                node_states[site].append((record.id,""))
                
    return node_states

def get_header_str(dict_values):
    header_str = ""
    for i in sorted(dict_values, key = lambda i : i[0]):
        header_str += f"{i[0]},"
    header_str = header_str.rstrip(",")
    return header_str
    
    
def find_what_sites_vary_unambiguously(node_states,state_differences):
    header_str = get_header_str(node_states["1"])
    
    with open(state_differences,"w") as fw:
        fw.write(f"site,{header_str}\n")

        for site in node_states:
            info = node_states[site]
            
            # get the set of unique bases at a given site
            count = set([i[1] for i in info if i[1]])
            
            #if there's more than one
            if len(count)>1:
                
                #needs to be kep consistent with header str
                info = sorted(info, key = lambda i : i[0])
                base_str = ""
                for i in info:
                    base_str += f"{i[1]},"
                    
                base_str = base_str.rstrip(",")
                fw.write(f"{site},{base_str}\n")
    
def load_unambiguous_varying_sites(infile):
    node_states_diff = collections.defaultdict(dict)
    with open(infile,"r") as f:
        reader = csv.DictReader(f)
        for row in reader:
            site = row["site"]
            for col in row:
                if col != "site":
                    node_states_diff[row["site"]][col] = row[col]
    return node_states_diff


def detect_tree_format(treefile, tree_format="auto"):
    if tree_format and tree_format != "auto":
        return tree_format
    lower = treefile.lower()
    if lower.endswith(".nex") or lower.endswith(".nexus"):
        return "nexus"
    try:
        with open(treefile, "r") as handle:
            head = handle.read(200).lower()
        if "#nexus" in head or "begin trees;" in head:
            return "nexus"
    except Exception:
        pass
    return "newick"


def load_tree(treefile, tree_format="auto"):
    fmt = detect_tree_format(treefile, tree_format)
    if fmt == "nexus":
        return bt.loadNexus(treefile, absoluteTime=False)
    return bt.loadNewick(treefile, absoluteTime=False)


def ensure_node_label(node):
    label = node.traits.get("label") if hasattr(node, "traits") else None
    if not label:
        label = getattr(node, "name", None)
    if label and "/" in label:
        label = label.split("/", 1)[0]
    if label:
        node.traits["label"] = label
    return label


def map_site_changes_to_branches(treefile, outfile,node_states,node_states_diff, tree_format="auto"): 
    my_tree=load_tree(treefile, tree_format=tree_format)
    last_node = ""
    current_node = ""

    with open(outfile,"w") as fw:
        fw.write("parent,child,site,snp,dimer\n")

        for k in my_tree.Objects:
            current_node = k
            ensure_node_label(current_node)

            if last_node:
                node_name = current_node.traits.get("label")
                parent_name = current_node.parent.traits.get("label")
                if not node_name or not parent_name:
                    last_node = current_node
                    continue
                snps = []
                for site in node_states_diff:
                    if node_name not in node_states_diff[site] or parent_name not in node_states_diff[site]:
                        continue
                    node_base = node_states_diff[site][node_name]
                    parent_base = node_states_diff[site][parent_name]

                    if node_base != parent_base:
                        if node_base in ["A","C","G","T"] and parent_base in ["A","C","G","T"]:
                            snp = f"{parent_base}->{node_base}"
                            snps.append(snp)
                            if snp == "G->A":
                                dimer_site = f"{int(site)+1}"
                                dimer_base = ""

                                for i in node_states[dimer_site]:
                                    if i[0] == parent_name:
                                        dimer_base = i[1]
                                dimer = f"{parent_base}{dimer_base}"
                            elif snp == "C->T":
                                dimer_site = f"{int(site)-1}"
                                dimer_base = ""

                                for i in node_states[dimer_site]:
                                    if i[0] == parent_name:
                                        dimer_base = i[1]
                                dimer = f"{dimer_base}{parent_base}"
                            else:
                                dimer = ""
                            fw.write(f"{parent_name},{node_name},{site},{snp},{dimer}\n")

            last_node = current_node

def read_in_branch_snps(branch_snps):
    branch_snps_dict = collections.defaultdict(list)
    with open(branch_snps,"r") as f:
        reader = csv.DictReader(f)
        for row in reader:
            branch = f"{row['parent']}_{row['child']}"
            branch_snps_dict[branch].append((row['site'],row['snp'],row['dimer'])) 
    return branch_snps_dict

def get_branch_snps_sites(branch_snps):
    all_snps = collections.Counter()
    branch_snps_dict = collections.defaultdict(list)
    with open(branch_snps,"r") as f:
        reader = csv.DictReader(f)
        for row in reader:
            all_snps[int(row["site"])] += 1

            branch_snps_dict[int(row['site'])].append([row['parent'],row['child'],row['snp'],row['dimer']])
    
    homoplasies = {}
    for k in all_snps:
        if all_snps[k] > 1:
            homoplasies[k] = all_snps[k]
            
    # print(len(homoplasies))
    # print(homoplasies)
    return branch_snps_dict,homoplasies
    
def get_acc_to_metadata_map(metadata):
    acc_dict = {}
    with open(metadata,"r") as f:
        reader = csv.DictReader(f)
        for row in reader:
            try:
                acc_dict[row["accession"]] = row
            except:
                try:
                    acc_dict[row["\ufeffaccession"]] = row
                except:
                    pass
                
    return acc_dict


def get_fig_height(alignment):
    seqs = SeqIO.index(alignment, "fasta")
    n_tips = len(seqs)
    max_label_len = 0
    for rec_id in seqs:
        if len(rec_id) > max_label_len:
            max_label_len = len(rec_id)

    # heuristic to keep labels legible without excessive height
    height = (n_tips * 0.25) + (max_label_len * 0.02)
    if height < 8:
        return 8
    if height > 60:
        return 60
    return height


def get_fig_size(treefile, tree_format="auto", alignment=None):
    """Estimate figure (width, height) for tree rendering.

    Width scales with genetic distance and label length; height scales with tip count.
    """
    my_tree = load_tree(treefile, tree_format=tree_format)

    tips = []
    for node in my_tree.Objects:
        if node.branchType == 'leaf':
            label = ensure_node_label(node)
            if label:
                tips.append(label)

    n_tips = len(tips)
    max_label_len = max([len(t) for t in tips], default=0)

    if alignment:
        try:
            seqs = SeqIO.index(alignment, "fasta")
            n_tips = max(n_tips, len(seqs))
            max_label_len = max(max_label_len, max((len(rec_id) for rec_id in seqs), default=0))
        except Exception:
            pass

    # height: driven by number of tips (vertical spacing)
    height = (n_tips * 0.25) + (max_label_len * 0.01)
    height = min(max(height, 8), 60)

    # width: driven by genetic distance (tree height) + label space
    tree_height = getattr(my_tree, "treeHeight", None)
    if tree_height is None:
        width = 25
    else:
        r2t = 200000 * tree_height
        width = (math.sqrt(r2t) * 3) if r2t < 200 else 25

    label_space = (max_label_len * 0.12) + 2
    width = max(width, label_space + (height * 0.1))
    width = min(max(width, 10), 80)

    return width, height


def make_reconstruction_tree_figure_w_labels(outfile,branch_snps,treefile,point_style,justification,w=None,h=None):
    from .plotting import make_reconstruction_tree_figure_w_labels as _plot
    return _plot(outfile, branch_snps, treefile, point_style, justification, w, h)
    
def generate_reconstruction_files(alignment, state_out, state_differences):
    
    node_states = get_node_states_all_sites(state_out,alignment)
        
    find_what_sites_vary_unambiguously(node_states,state_differences)

    return node_states
    
def load_info(directory, alignment, treefile, state_out, state_differences, branch_snps_out, treefigureout,point_style,point_justify, node_states="",width=None,height=None):
    
    if not node_states:
        node_states = get_node_states_all_sites(state_out, alignment)

    node_states_diff = load_unambiguous_varying_sites(state_differences)

    map_site_changes_to_branches(treefile,
                                 branch_snps_out,
                                 node_states,
                                 node_states_diff)

    make_reconstruction_tree_figure_w_labels(treefigureout,
                                    branch_snps_out,
                                    treefile,
                                    point_style,
                                    point_justify,
                                    width,
                                    height)
    
def get_gene_boundaries(gene_boundaries_file):
    genes = {}
    gene_id = 0
    with open(gene_boundaries_file,"r") as f:
        reader = csv.DictReader(f)
        for row in reader:
            gene_id +=1
            name = f"{row['Name'].replace(' ','_')}_{gene_id}"
            start = int(row["Minimum"])
            end = int(row["Maximum"]) + 1
            length = int(row["Length"])
            direction = row["Direction"]
            genes[(start,end)] = (name,length,direction)
    return genes

def get_grantham_scores(grantham_scores_file):
    grantham_scores = {}

    with open(grantham_scores_file,"r") as f:
        reader = csv.DictReader(f, delimiter="\t")
        for row in reader:
            for col in row:
                if col!="FIRST":
                    mutation = f"{row['FIRST']}{col}"

                    if row[col] != "0":
                        grantham_scores[mutation] = int(row[col])                    


#!/usr/bin/env python3
"""Phylogenetic QC subcommand wrapper."""
import os
import logging


def main(args):
    """Run phylogenetic QC.

    The implementation defers heavy imports until called so tests can import the package without needing all scientific deps.
    """
    logging.info("Starting phylogenetic QC")
    if not hasattr(args, 'phylogeny'):
        raise ValueError("Expected argparse Namespace from raccoon.command.build_parser")

    try:
        # Lazy import
        from raccoon.utils import phylo_functions as pf
        from raccoon.utils import io
        from raccoon.utils.constants import KEY_OUTDIR, KEY_OUTFILENAME, KEY_PHYLOGENY, KEY_RUN_APOBEC3_PHYLO

        outdir = args.outdir or os.getcwd()
        
        # validate output directory
        if not io.ensure_output_directory(outdir):
            return 1
        
        # validate input files
        assembly_refs = getattr(args, 'assembly_refs', None)
        if assembly_refs and not io.validate_alignment_file(assembly_refs):
            return 1

        treefile = io.resolve_existing_file(getattr(args, 'phylogeny', None), outdir, "Phylogeny file")
        if not treefile:
            return 1
        
        mask_file = getattr(args, 'mask_file', None)
        if mask_file and not io.validate_input_file(mask_file, "Mask file"):
            return 1

        alignment = getattr(args, 'alignment', None)
        if alignment and not io.validate_alignment_file(alignment):
            return 1

        state_file = io.resolve_asr_state_file(getattr(args, 'asr_state', None), treefile)
        if getattr(args, 'asr_state', None) and not state_file:
            return 1
        
        config = {}
        config[KEY_OUTDIR] = outdir
        phylogeny_base = os.path.splitext(os.path.basename(treefile))[0]
        config[KEY_OUTFILENAME] = phylogeny_base
        config[KEY_PHYLOGENY] = phylogeny_base
        config[KEY_RUN_APOBEC3_PHYLO] = args.run_apobec

        outgroup_ids = []
        if args.outgroup_ids:
            outgroup_ids = [x.strip() for x in args.outgroup_ids.split(',') if x.strip()]

        mask_file = mask_file or os.path.join(outdir, f"{phylogeny_base}.mask.csv")

        flags_csv = pf.run_phylo_qc(
            treefile=treefile,
            tree_format=args.tree_format,
            outdir=outdir,
            alignment=alignment,
            state_file=state_file,
            assembly_refs=assembly_refs,
            long_branch_sd=args.long_branch_sd,
            include_apobec=args.run_apobec,
            include_adar=args.run_adar,
            adar_window=args.adar_window,
            adar_min_count=args.adar_min_count,
        )
        try:
            from raccoon.utils import reporting
            reporting.generate_phylo_report(
                outdir=outdir,
                treefile=treefile,
                flags_csv=flags_csv,
                tree_format=args.tree_format,
            )
        except Exception:
            logging.exception("Failed to generate phylo report")
        logging.info("Phylogenetic QC finished")
        return 0
    except Exception:
        logging.exception("Phylo QC failed")
        return 2



#!/usr/bin/env python3
"""Combine FASTA files and optionally harmonise headers from metadata."""
import csv
import logging
import os
import sys
from typing import Dict, Iterable, Optional

from Bio import SeqIO


def get_field(row: Dict[str, str], field: str) -> str:
    if field in row:
        return row.get(field, "") or ""
    bom_field = f"\ufeff{field}"
    return row.get(bom_field, "") or ""


def load_metadata_map(metadata_path: str, id_field: str, delimiter: str) -> Dict[str, Dict[str, str]]:
    metadata = {}
    with open(metadata_path, "r", newline="") as handle:
        reader = csv.DictReader(handle, delimiter=delimiter)
        for row in reader:
            key = get_field(row, id_field)
            if not key:
                continue
            metadata[key] = row
    return metadata


def _infer_delimiter(path: str, default: str) -> str:
    lowered = path.lower()
    if lowered.endswith(".tsv") or lowered.endswith(".tab"):
        return "\t"
    return default


def load_metadata_maps(metadata_paths: Iterable[str], id_field: str, delimiter: str) -> Dict[str, Dict[str, str]]:
    merged = {}
    for path in metadata_paths:
        effective_delimiter = _infer_delimiter(path, delimiter)
        merged.update(load_metadata_map(path, id_field, effective_delimiter))
    return merged


def format_header(
    record_id: str,
    row: Dict[str, str],
    location_field: str,
    date_field: str,
    sep: str,
) -> str:
    def _sanitize(value: str) -> str:
        if value is None:
            return ""
        cleaned = str(value).strip()
        return "_".join(cleaned.split())

    location = _sanitize(get_field(row, location_field))
    date = _sanitize(get_field(row, date_field))
    return f"{record_id}{sep}{location}{sep}{date}"


def write_fasta_record(handle, header: str, sequence: str) -> None:
    sequence = "".join(sequence.split()).upper()
    handle.write(f">{header}\n{sequence}\n")


def n_content(seq: str) -> float:
    seq = seq.upper()
    if not seq:
        return 0.0
    return seq.count("N") / len(seq)


def parse_record_id(record_id: str, delimiter: str, field_index: int) -> str:
    if field_index < 0:
        return record_id
    parts = record_id.split(delimiter) if delimiter else [record_id]
    if not parts:
        return record_id
    if field_index >= len(parts):
        return record_id
    return parts[field_index] or record_id


def main(args):
    """Combine fasta files into a single upper-case, unwrapped FASTA."""
    if not hasattr(args, "inputs"):
        raise ValueError("Expected argparse Namespace from raccoon.command.build_parser")

    try:
        from raccoon.utils import io

        inputs = args.inputs or []
        if not inputs:
            logging.error("No input FASTA files provided")
            return 1

        for path in inputs:
            if not io.validate_input_file(path, "FASTA file"):
                return 1

        metadata_map: Optional[Dict[str, Dict[str, str]]] = None
        metadata_paths = list(getattr(args, "metadata", []) or [])
        metadata_id_field = getattr(args, "metadata_id_field", "id")
        metadata_location_field = getattr(args, "metadata_location_field", "location")
        metadata_date_field = getattr(args, "metadata_date_field", "date")
        metadata_delimiter = getattr(args, "metadata_delimiter", ",")
        header_separator = getattr(args, "header_separator", "|")
        id_delimiter = getattr(args, "id_delimiter", "|")
        id_field = getattr(args, "id_field", 0)
        min_length = getattr(args, "min_length", None)
        max_n_content = getattr(args, "max_n_content", None)

        if metadata_paths:
            for path in metadata_paths:
                if not io.validate_input_file(path, "Metadata file"):
                    return 1
            metadata_map = load_metadata_maps(metadata_paths, metadata_id_field, metadata_delimiter)

        output_path = args.output or "combined.fasta"
        if output_path == "-":
            out_handle = sys.stdout
            close_handle = False
        else:
            if not io.ensure_parent_directory(output_path):
                return 1
            out_handle = open(output_path, "w")
            close_handle = True

        filtered_count = 0
        kept_count = 0
        filter_failures = []
        metadata_issues = []
        try:
            for path in inputs:
                for rec in SeqIO.parse(path, "fasta"):
                    seq = str(rec.seq)
                    seq_len = len(seq)
                    n_prop = n_content(seq)
                    parsed_id = parse_record_id(rec.id, id_delimiter, id_field)
                    metadata_row = None
                    location_value = ""
                    date_value = ""
                    if metadata_map is not None:
                        metadata_row = metadata_map.get(parsed_id)
                        if metadata_row:
                            location_value = get_field(metadata_row, metadata_location_field).strip()
                            date_value = get_field(metadata_row, metadata_date_field).strip()
                        else:
                            logging.warning("No metadata row found for %s", parsed_id)
                    reasons = []
                    if min_length is not None and seq_len < min_length:
                        reasons.append(f"length < {min_length}")
                    if max_n_content is not None and n_prop > max_n_content:
                        reasons.append(f"N content > {max_n_content}")
                    status = "filtered" if reasons else "kept"
                    if metadata_map is not None:
                        if metadata_row:
                            if not location_value:
                                metadata_issues.append({
                                    "file": os.path.basename(path),
                                    "id": rec.id,
                                    "parsed_id": parsed_id,
                                    "status": status,
                                    "issue": "missing location",
                                    "location": location_value,
                                    "date": date_value,
                                })
                            if not date_value:
                                metadata_issues.append({
                                    "file": os.path.basename(path),
                                    "id": rec.id,
                                    "parsed_id": parsed_id,
                                    "status": status,
                                    "issue": "missing date",
                                    "location": location_value,
                                    "date": date_value,
                                })
                        else:
                            metadata_issues.append({
                                "file": os.path.basename(path),
                                "id": rec.id,
                                "parsed_id": parsed_id,
                                "status": status,
                                "issue": "missing metadata row",
                                "location": "",
                                "date": "",
                            })
                    if reasons:
                        filter_failures.append({
                            "file": os.path.basename(path),
                            "id": rec.id,
                            "parsed_id": parsed_id,
                            "length": seq_len,
                            "n_content": round(n_prop, 6),
                            "reason": "; ".join(reasons),
                        })
                        filtered_count += 1
                        continue
                    header = rec.id
                    if metadata_map is not None:
                        row = metadata_map.get(parsed_id)
                        if row:
                            header = format_header(
                                parsed_id,
                                row,
                                metadata_location_field,
                                metadata_date_field,
                                header_separator,
                            )
                    write_fasta_record(out_handle, header, seq)
                    kept_count += 1
        finally:
            if close_handle:
                out_handle.close()

        report_outdir = os.path.dirname(output_path) or os.getcwd()
        if not io.ensure_output_directory(report_outdir):
            return 1
        if min_length is not None or max_n_content is not None:
            filter_csv = os.path.join(report_outdir, "seq_qc_filter_failures.csv")
            with open(filter_csv, "w", newline="") as handle:
                writer = csv.DictWriter(
                    handle,
                    fieldnames=["file", "id", "parsed_id", "length", "n_content", "reason"],
                    lineterminator="\n",
                )
                writer.writeheader()
                writer.writerows(filter_failures)
        if metadata_map is not None:
            metadata_csv = os.path.join(report_outdir, "seq_qc_metadata_issues.csv")
            with open(metadata_csv, "w", newline="") as handle:
                writer = csv.DictWriter(
                    handle,
                    fieldnames=["file", "id", "parsed_id", "status", "issue", "location", "date"],
                    lineterminator="\n",
                )
                writer.writeheader()
                writer.writerows(metadata_issues)

        try:
            from raccoon.utils import reporting
            reporting.generate_combine_report(
                outdir=report_outdir,
                output_fasta=output_path if output_path != "-" else "",
                input_fastas=inputs,
                metadata_paths=metadata_paths or None,
                metadata_id_field=metadata_id_field,
                metadata_location_field=metadata_location_field,
                metadata_date_field=metadata_date_field,
                header_separator=header_separator,
                min_length=min_length,
                max_n_content=max_n_content,
                filter_failures=filter_failures,
                metadata_issues=metadata_issues,
            )
        except Exception:
            logging.exception("Failed to generate combine report")

        logging.info("Combined %d input FASTA files", len(inputs))
        if min_length is not None or max_n_content is not None:
            logging.info("Filtered %d sequences; kept %d", filtered_count, kept_count)
        return 0
    except Exception:
        logging.exception("Combine failed")
        return 2



"""Plotly-based interactive tree rendering utilities.
This is used for tree visualization in the phylo-qc report and may be useful for other applications.
The code was inspired by the baltic library by Gytis Dudas (https://github.com/evogytis/baltic)
"""
from __future__ import annotations

from typing import Any, Dict, List, Optional

import csv
import os

import plotly.graph_objects as go
import plotly.io as pio

from .reconstruction_functions import load_tree, ensure_node_label


def _scale_values(values: List[Optional[float]], factor: float) -> List[Optional[float]]:
    return [None if v is None else v * factor for v in values]


def _apply_plot_style(fig: go.Figure) -> None:
    fig.update_layout(
        font=dict(family="Helvetica Neue", size=12, color="#111"),
        colorway=["#4BA3A8", "#7A6BB1", "#D08BA8"],
        plot_bgcolor="white",
        paper_bgcolor="white",
        showlegend=True,
        hoverlabel=dict(font=dict(color="#ffffff"), bordercolor="#ffffff"),
    )
    fig.update_xaxes(showline=True, linecolor="black", linewidth=1, gridcolor="rgba(0,0,0,0.05)")
    fig.update_yaxes(showline=True, linecolor="black", linewidth=1, gridcolor="rgba(0,0,0,0.05)")


def _plot_div(fig: go.Figure, div_id: Optional[str] = None) -> str:
    return pio.to_html(
        fig,
        include_plotlyjs="cdn",
        full_html=False,
        config={"displayModeBar": False, "showTips": False, "doubleClick": False, "responsive": True},
        auto_play=False,
        div_id=div_id,
    )


def _format_traits(traits: Dict[str, Any]) -> str:
    lines = []
    for key in sorted(traits):
        value = traits.get(key)
        if value is None or value == "":
            continue
        lines.append(f"{key}: {value}")
    return "<br>".join(lines)


def _read_branch_snps(branch_snps_path: Optional[str]) -> Dict[str, List[Dict[str, str]]]:
    if not branch_snps_path or not os.path.exists(branch_snps_path):
        return {}
    branch_snps: Dict[str, List[Dict[str, str]]] = {}
    with open(branch_snps_path, "r") as handle:
        reader = csv.DictReader(handle)
        for row in reader:
            parent = row.get("parent", "")
            child = row.get("child", "")
            if not parent or not child:
                continue
            branch_key = f"{parent}_{child}"
            branch_snps.setdefault(branch_key, []).append({
                "site": row.get("site", ""),
                "snp": row.get("snp", ""),
                "dimer": row.get("dimer", ""),
            })
    return branch_snps


def _format_branch_snps(branch_key: str, snps: List[Dict[str, str]], max_rows: int = 20) -> str:
    if not snps:
        return f"Branch: {branch_key}<br>No mutations"
    lines = [f"Branch: {branch_key}", f"Mutations: {len(snps)}"]
    for item in snps[:max_rows]:
        site = item.get("site", "")
        snp = item.get("snp", "")
        dimer = item.get("dimer", "")
        suffix = f" ({dimer})" if dimer else ""
        lines.append(f"{site}: {snp}{suffix}")
    if len(snps) > max_rows:
        lines.append(f"… and {len(snps) - max_rows} more")
    return "<br>".join(lines)


def _parse_label_traits(label: str) -> Dict[str, Any]:
    traits: Dict[str, Any] = {}
    if not label:
        return traits
    parts = [p.strip() for p in label.split("|")]
    if len(parts) >= 2 and parts[1]:
        traits.setdefault("location", parts[1])
    if len(parts) >= 3 and parts[2]:
        traits.setdefault("date", parts[2])
    return traits


def build_tree_plot(
    treefile: str,
    tree_format: str = "auto",
    branch_snps_path: Optional[str] = None,
) -> str:
    try:
        tree = load_tree(treefile, tree_format=tree_format)
    except Exception:
        return ""

    try:
        tree.drawTree()
    except Exception:
        return ""

    line_x: List[Optional[float]] = []
    line_y: List[Optional[float]] = []

    try:
        internal_nodes = tree.getInternal()
    except Exception:
        internal_nodes = []

    for node in internal_nodes:
        if not getattr(node, "children", None):
            continue
        child_ys = []
        for child in node.children:
            if child.x is None or child.y is None or node.x is None:
                continue
            line_x.extend([node.x, child.x, None])
            line_y.extend([child.y, child.y, None])
            child_ys.append(child.y)
        if child_ys and node.x is not None:
            line_x.extend([node.x, node.x, None])
            line_y.extend([min(child_ys), max(child_ys), None])

    tip_x: List[float] = []
    tip_y: List[float] = []
    tip_hover: List[str] = []
    tip_traits: List[Dict[str, Any]] = []
    tip_labels: List[str] = []

    node_x: List[float] = []
    node_y: List[float] = []
    node_hover: List[str] = []

    branch_hover_x: List[float] = []
    branch_hover_y: List[float] = []
    branch_hover_text: List[str] = []

    branch_snps = _read_branch_snps(branch_snps_path)

    for obj in getattr(tree, "Objects", []):
        label = ensure_node_label(obj) or getattr(obj, "name", "") or ""
        traits = dict(getattr(obj, "traits", {}) or {})
        traits.setdefault("label", label)
        traits.setdefault("branch_type", getattr(obj, "branchType", ""))
        traits.update(_parse_label_traits(label))

        if getattr(obj, "is_leaflike", lambda: False)():
            if obj.x is None or obj.y is None:
                continue
            tip_x.append(obj.x)
            tip_y.append(obj.y)
            tip_traits.append(traits)
            tip_hover.append(_format_traits(traits))
            tip_labels.append(label)
        elif getattr(obj, "is_node", lambda: False)():
            if obj.x is None or obj.y is None:
                continue
            node_x.append(obj.x)
            node_y.append(obj.y)
            node_hover.append(_format_traits(traits))

    for node in internal_nodes:
        parent_label = ensure_node_label(node) or getattr(node, "name", "") or ""
        for child in getattr(node, "children", []) or []:
            if child.x is None or child.y is None or node.x is None:
                continue
            child_label = ensure_node_label(child) or getattr(child, "name", "") or ""
            branch_key = f"{parent_label}_{child_label}"
            snps = branch_snps.get(branch_key)
            if not snps:
                continue
            branch_hover_x.append((node.x + child.x) / 2)
            branch_hover_y.append(child.y)
            branch_hover_text.append(_format_branch_snps(branch_key, snps))

    color_keys = set()
    for traits in tip_traits:
        for key, value in traits.items():
            if value is not None and value != "":
                color_keys.add(key)
    preferred = ["location", "date", "branch_type"]
    color_keys = preferred + sorted([k for k in color_keys if k not in preferred])
    default_key = color_keys[0] if color_keys else "branch_type"

    palette = [
        "#4BA3A8",
        "#7A6BB1",
        "#D08BA8",
        "#5A9BD4",
        "#B9A44C",
        "#8BC59A",
        "#E26D5C",
        "#6C8EA5",
    ]

    color_maps: Dict[str, Dict[str, str]] = {}
    for key in color_keys:
        values = sorted({str(t.get(key, "")) for t in tip_traits if t.get(key, "") != ""})
        mapping = {}
        for idx, value in enumerate(values):
            mapping[value] = palette[idx % len(palette)]
        color_maps[key] = mapping

    def _colors_for(key: str) -> List[str]:
        mapping = color_maps.get(key, {})
        return [mapping.get(str(t.get(key, "")), "#4d4d4d") for t in tip_traits]

    x_values = [v for v in line_x if v is not None]
    x_values.extend(tip_x)
    x_values.extend(node_x)
    x_min = min(x_values) if x_values else 0.0
    x_max = max(x_values) if x_values else 1.0
    x_pad = (x_max - x_min) * 0.02 if x_values else 0.1
    x_min_pad = x_min - x_pad
    x_max_pad = x_max + x_pad

    fig = go.Figure()
    fig.add_trace(go.Scatter(
        x=line_x,
        y=line_y,
        mode="lines",
        line=dict(color="#222", width=1),
        hoverinfo="skip",
        showlegend=False,
        name="Branches",
    ))
    fig.add_trace(go.Scatter(
        x=tip_x,
        y=tip_y,
        mode="markers",
        marker=dict(size=8, color=_colors_for(default_key)),
        hovertemplate="%{text}<extra></extra>",
        text=tip_hover,
        name="Tips",
    ))
    fig.add_trace(go.Scatter(
        x=node_x,
        y=node_y,
        mode="markers",
        marker=dict(size=4, color="#999999"),
        hovertemplate="%{text}<extra></extra>",
        text=node_hover,
        name="Nodes",
    ))
    if tip_x:
        tip_span = max(tip_x) - min(tip_x)
        x_offset = tip_span * 0.01 if tip_span else 0.01
        tip_label_x = [x + x_offset for x in tip_x]
    else:
        tip_label_x = []

    label_max_x = max(tip_label_x) if tip_label_x else None
    if label_max_x is not None:
        label_span = (x_max - x_min) if x_values else 1.0
        max_label_len = max((len(label) for label in tip_labels), default=0)
        label_extra = label_span * max_label_len * 0.006
        if label_max_x + label_extra > x_max_pad:
            x_max_pad = label_max_x + label_extra + x_pad
    x_range_default = (x_min_pad, x_max_pad)

    fig.add_trace(go.Scatter(
        x=tip_label_x,
        y=tip_y,
        mode="text",
        text=tip_labels,
        textposition="middle right",
        textfont=dict(size=10, color="#222"),
        hoverinfo="skip",
        showlegend=False,
        name="Tip labels",
        visible=True,
    ))
    fig.add_trace(go.Scatter(
        x=branch_hover_x,
        y=branch_hover_y,
        mode="markers",
        marker=dict(size=10, color="rgba(0,0,0,0)"),
        hovertemplate="%{text}<extra></extra>",
        text=branch_hover_text,
        showlegend=False,
        name="Branch mutations",
    ))

    scale_values = [round(1.0 + (i * 0.01), 2) for i in range(101)]
    y_min = min(tip_y) if tip_y else 0
    y_max = max(tip_y) if tip_y else 0
    y_span = y_max - y_min
    y_pad = max(1.0, y_span * 0.03) if tip_y else 1.0
    y_min_pad = y_min - y_pad
    y_max_pad = y_max + y_pad
    base_height = max(500, int((y_max_pad if tip_y else 0) * 15)) if tip_y else 500
    y_range_default = (y_min_pad, y_max_pad)
    height_default = base_height

    slider_steps = []
    for scale in scale_values:
        slider_steps.append({
            "label": "",
            "method": "relayout",
            "args": [{
                "height": int(base_height * scale),
                "yaxis.range": list(y_range_default),
            }],
        })

    color_buttons = []
    for key in color_keys:
        color_buttons.append({
            "label": key,
            "method": "restyle",
            "args": [{"marker.color": [_colors_for(key)]}, [1]],
        })

    fig.update_layout(
        xaxis_title="Substitutions per site",
        showlegend=False,
        height=height_default,
        xaxis=dict(range=list(x_range_default)),
        yaxis=dict(range=list(y_range_default) if tip_y else None),
        sliders=[{
            "active": 0,
            "pad": {"t": 0, "b": 0},
            "ticklen": 0,
            "tickwidth": 0,
            "len": 0.28,
            "x": 0.62,
            "y": 1.12,
            "xanchor": "left",
            "yanchor": "top",
            "steps": slider_steps,
        }],
        updatemenus=[{
            "buttons": color_buttons,
            "direction": "down",
            "showactive": True,
            "x": 0.12,
            "y": 1.12,
            "xanchor": "left",
            "yanchor": "top",
        }, {
            "buttons": [
                {"label": "On", "method": "restyle", "args": [{"visible": True}, [3]]},
                {"label": "Off", "method": "restyle", "args": [{"visible": False}, [3]]},
            ],
            "direction": "down",
            "showactive": False,
            "x": 0.36,
            "y": 1.12,
            "xanchor": "left",
            "yanchor": "top",
        }, {
            "buttons": [
                {
                    "label": "Reset view",
                    "method": "relayout",
                    "args": [{
                        "xaxis.range[0]": x_range_default[0],
                        "xaxis.range[1]": x_range_default[1],
                        "yaxis.autorange": False,
                        "yaxis.range[0]": y_range_default[0],
                        "yaxis.range[1]": y_range_default[1],
                        "height": height_default,
                        "sliders[0].active": 0,
                    }],
                }
            ],
            "type": "buttons",
            "direction": "right",
            "showactive": False,
            "x": 0.98,
            "y": 1.12,
            "xanchor": "right",
            "yanchor": "top",
        }],
        annotations=[
            {
                "text": "Colour by",
                "xref": "paper",
                "yref": "paper",
                "x": 0.02,
                "y": 1.12,
                "showarrow": False,
                "xanchor": "left",
                "yanchor": "top",
                "font": {"size": 12, "color": "#111"},
            },
            {
                "text": "Tip labels",
                "xref": "paper",
                "yref": "paper",
                "x": 0.28,
                "y": 1.12,
                "showarrow": False,
                "xanchor": "left",
                "yanchor": "top",
                "font": {"size": 12, "color": "#111"},
            },
            {
                "text": "Expand tree",
                "xref": "paper",
                "yref": "paper",
                "x": 0.56,
                "y": 1.12,
                "showarrow": False,
                "xanchor": "left",
                "yanchor": "top",
                "font": {"size": 12, "color": "#111"},
            },
        ],
        margin=dict(l=60, r=20, t=60, b=40),
    )
    _apply_plot_style(fig)
    fig.update_yaxes(showticklabels=False, ticks="", title=None, showline=False)
    return _plot_div(fig)



# Constants for CSV/field names and notes to avoid typos
KEY_NAME = "Name"
KEY_MINIMUM = "Minimum"
KEY_MAXIMUM = "Maximum"
KEY_LENGTH = "Length"
KEY_PRESENT_IN = "present_in"
KEY_NOTE = "note"
KEY_SITES_TO_MASK = "sites_to_mask"
KEY_MASK_FILE = "mask_file"
KEY_ISSUES_FOUND = "issues_found"
KEY_CLUSTERED_SNP_COUNT = "clustered_snp_count"

# Note values
NOTE_CLUSTERED_SNPS = "clustered_snps"
NOTE_N_ADJACENT = "N_adjacent"
NOTE_GAP_ADJACENT = "gap_adjacent"
NOTE_REVERSION = "reversion"
NOTE_CONVERGENT = "convergent_snp"
NOTE_FRAME_BREAK = "frame_break"
NOTE_UNIQUE_SNP = "unique_snp"

# General config keys (migrated from utils/config.py)
KEY_INPUT_FASTA = "fasta"
KEY_REFERENCE_FASTA = "reference_fasta"
KEY_TO_MASK = "to_mask"
KEY_GENE_BOUNDARIES = "gene_boundaries"
KEY_OUTDIR = "outdir"
KEY_EXCLUDE_FILE = "exclude_file"
KEY_OUTFILE_STEM = "outfile_stem"
KEY_OUTFILENAME = "outfilename"
KEY_VERBOSE = "verbose"
KEY_THREADS = "threads"
KEY_PHYLO_THREADS = "phylo_threads"
KEY_NO_MASK = "no_mask"
KEY_SEQUENCE_MASK = "sequence_mask"
KEY_TRIM_END = "trim_end"
KEY_EXTRACT_CDS = "extract_cds"
KEY_SEQ_QC = "seq_qc"
KEY_ASSEMBLY_REFERENCES = "assembly_references"

KEY_OUTGROUPS = "outgroups"
KEY_PHYLOGENY = "phylogeny"
KEY_PHYLOGENY_SVG = "phylogeny_svg"
KEY_OUTGROUP_STRING = "outgroup_string"
KEY_GRANTHAM_SCORES = "grantham_scores"

KEY_TREE = "tree"
KEY_BRANCH_RECONSTRUCTION = "branch_reconstruction"
KEY_ASR_TREE = "asr_tree"
KEY_ASR_STATE = "asr_state"
KEY_ASR_ALIGNMENT = "asr_alignment"

KEY_FIG_HEIGHT = "fig_height"
KEY_FIG_WIDTH = "fig_width"
KEY_POINT_STYLE = "point_style"
KEY_POINT_JUSTIFY = "point_justify"

# Additional feature flags used in code
KEY_RUN_APOBEC3_PHYLO = "run_apobec3_phylo"

KEY_GENBANK = "genbank"
KEY_REFERENCE_ID = "reference_id"
KEY_N_THRESHOLD = "n_threshold"
KEY_CLUSTER_WINDOW = "cluster_window"
KEY_CLUSTER_COUNT = "cluster_count"
KEY_MASK_CLUSTERED = "mask_clustered"
KEY_MASK_N_ADJACENT = "mask_n_adjacent"
KEY_MASK_GAP_ADJACENT = "mask_gap_adjacent"
KEY_MASK_FRAME_BREAK = "mask_frame_break"

DEFAULT_N_THRESHOLD = 0.2
DEFAULT_CLUSTER_WINDOW = 10
DEFAULT_CLUSTER_COUNT = 3



#!/usr/bin/env python3
import sys
import os
import logging
from raccoon.utils.misc import green, cyan

import select
from Bio import SeqIO
from Bio import AlignIO
import collections
import csv
import math
import re

from .constants import *

def sliding_window(elements, window_size):
    """Yield sliding windows over elements."""
    if len(elements) <= window_size:
        yield elements
        return
    for i in range(len(elements) - window_size + 1):
        yield elements[i:i+window_size]


def read_alignment(path, fmt='fasta'):
    """Read an alignment file using AlignIO and return a MultipleSeqAlignment."""
    with open(path, 'r') as fh:
        aln = AlignIO.read(fh, fmt)
    return aln


def find_high_N_sequences(aln, threshold):
    """Find sequences with N content greater than threshold.

    Returns a list of tuples (record_id, n_content_float).
    """
    flagged = []
    for rec in aln:
        seq = str(rec.seq).upper()
        n_count = seq.count('N')
        n_content = n_count / len(seq) if len(seq) else 0.0
        if n_content > threshold:
            flagged.append((rec.id, round(n_content, 3)))
    return flagged


def find_frame_breaking_indels(aln, genbank_path, reference_id=None):
    """Find frame-breaking indels using coding coordinates from a GenBank file.

    This function expects the alignment to contain the reference sequence with
    id matching `reference_id`. It returns a dict of sites (1-based positions)
    flagged due to frame-breaking indels in coding sequences.
    """
    if not genbank_path:
        return {}

    # load genbank
    try:
        gb = None
        for rec in SeqIO.parse(genbank_path, 'genbank'):
            if reference_id is None or rec.id == reference_id or rec.name == reference_id:
                gb = rec
                break
        if gb is None:
            raise FileNotFoundError(f"Reference {reference_id} not found in GenBank file")
    except Exception as exc:
        logging.error("Error reading genbank file: %s", exc)
        raise

    # build mapping from reference ungapped index -> alignment column index
    ref_rec = None
    for rec in aln:
        if rec.id == gb.id or rec.id == gb.name or (reference_id and rec.id == reference_id):
            ref_rec = rec
            break
    if ref_rec is None:
        raise ValueError("Reference sequence not found in alignment")

    ref_to_col = {}
    ungapped_pos = 0
    for col_idx in range(aln.get_alignment_length()):
        base = ref_rec.seq[col_idx]
        if base != '-':
            ungapped_pos += 1
            ref_to_col[ungapped_pos] = col_idx

    sites_to_mask = {}
    # iterate CDS features
    for feat in gb.features:
        if feat.type != 'CDS':
            continue
        # location using feature.location: start/end are 0-based on unaligned ref
        start = int(feat.location.start) + 1
        end = int(feat.location.end)
        # map to alignment columns
        cols = [ref_to_col.get(pos) for pos in range(start, end + 1)]
        cols = [c for c in cols if c is not None]
        if not cols:
            continue
        # for each sequence, if CDS length is not multiple of 3, flag indel sites only
        for rec in aln:
            non_gaps = sum(1 for c in cols if rec.seq[c] != '-')
            if non_gaps % 3 != 0:
                for c in cols:
                    if rec.seq[c] != '-':
                        continue
                    site = c + 1
                    if site not in sites_to_mask:
                        sites_to_mask[site] = {
                            'Name': site,
                            'Minimum': site,
                            'Maximum': site,
                            'Length': 1,
                            'present_in': [rec.id],
                            'note': set(['frame_break'])
                        }
                    else:
                        sites_to_mask[site]['present_in'].append(rec.id)
                        sites_to_mask[site]['note'].add('frame_break')
    return sites_to_mask


def _safe_int(value):
    try:
        return int(str(value).strip())
    except Exception:
        return None


def parse_mask_sites(mask_file):
    if not mask_file or not os.path.exists(mask_file):
        return set()

    positions = set()
    with open(mask_file, "r") as handle:
        reader = csv.DictReader(handle)
        for row in reader:
            entry_type = (row.get("type") or "site").strip().lower()
            if entry_type == "sequence_record":
                continue

            min_val = _safe_int(row.get(KEY_MINIMUM))
            max_val = _safe_int(row.get(KEY_MAXIMUM))

            if min_val is None or max_val is None:
                name_val = row.get(KEY_NAME, "") or row.get("flagged", "") or ""
                digits = re.findall(r"\d+", name_val)
                if digits:
                    min_val = int(digits[0])
                    max_val = int(digits[1]) if len(digits) > 1 else min_val

            if min_val is None or max_val is None:
                continue

            if max_val < min_val:
                min_val, max_val = max_val, min_val

            for pos in range(min_val, max_val + 1):
                positions.add(pos)

    return positions


def parse_mask_rows(mask_file):
    if not mask_file or not os.path.exists(mask_file):
        return [], []

    site_positions = set()
    sequences_to_remove = set()
    with open(mask_file, "r") as handle:
        reader = csv.DictReader(handle)
        for row in reader:
            entry_type = (row.get("type") or "site").strip().lower()
            flagged = (row.get("flagged") or row.get(KEY_NAME) or "").strip()
            if entry_type == "sequence_record":
                if flagged:
                    sequences_to_remove.add(flagged)
                continue

            min_val = _safe_int(row.get(KEY_MINIMUM))
            max_val = _safe_int(row.get(KEY_MAXIMUM))
            if min_val is None or max_val is None:
                digits = re.findall(r"\d+", flagged)
                if digits:
                    min_val = int(digits[0])
                    max_val = int(digits[1]) if len(digits) > 1 else min_val

            if min_val is None or max_val is None:
                continue

            if max_val < min_val:
                min_val, max_val = max_val, min_val

            for pos in range(min_val, max_val + 1):
                site_positions.add(pos)

    return sorted(site_positions), sorted(sequences_to_remove)


def apply_mask_to_alignment(alignment_path, mask_file, output_path, mask_char="N"):
    aln = read_alignment(alignment_path)
    positions, sequences_to_remove = parse_mask_rows(mask_file)

    if sequences_to_remove:
        aln = type(aln)([rec for rec in aln if rec.id not in sequences_to_remove])

    if not positions:
        with open(output_path, "w") as handle:
            AlignIO.write(aln, handle, "fasta")
        return 0

    max_len = aln.get_alignment_length()
    valid_positions = [pos for pos in positions if 1 <= pos <= max_len]

    for rec in aln:
        seq_list = list(str(rec.seq))
        for pos in valid_positions:
            seq_list[pos - 1] = mask_char
        rec.seq = type(rec.seq)("".join(seq_list))

    with open(output_path, "w") as handle:
        AlignIO.write(aln, handle, "fasta")
    return len(valid_positions)


def analyze_alignment(alignment, n_window=2, gap_window=1, snp_window=10, snp_count=3):
    """Analyze an alignment and return helper structures used by QC checks.

    Returns a tuple (aln, aln_len, majority, unique_mutations, snps_near_n, snps_near_gap)
    """
    
    if isinstance(alignment, str):
        aln = AlignIO.read(alignment, 'fasta')
    else:
        aln = alignment
    aln_len = aln.get_alignment_length()

    bases = set(["A", "T", "G", "C"]) 
    # # get majority base per site
    # majority = [None] * aln_len
    # for i in range(aln_len):
    #     col = [str(rec.seq[i]).upper() for rec in aln if str(rec.seq[i]).upper() in bases]
    #     if col:
    #         c = collections.Counter(col)
    #         majority[i] = c.most_common(1)[0][0]

    # set up dicts keyed by sequence id and values a set of indexes
    unique_mutations = collections.defaultdict(set)
    snps_near_n = collections.defaultdict(set)
    snps_near_gap = collections.defaultdict(set)
    clustered_snps = collections.defaultdict(set)
    
    # the assumption here is that there are relatively few variable sites, 
    # so we can iterate column-wise and skip most sites quickly, then do more detailed checks only on the variable sites.
    
    snp_cols = set()
    for i in range(aln_len):
        
        col = set()
        
        for s in aln:

            #find the columns with variable sites
            base = s.seq[i].upper()
            if base in bases:
                col.add(base)
                
        if col:
            if len(col)>1:
                #do this for only the variable sites to save time
                snp_cols.add(i)
                    
                col_dict = collections.defaultdict(list) # key is base, value is list of sequence ids with that base at this position
                col_counter = collections.Counter() # key is base, value is count of sequences with that base at this position
                
                for s in aln:
                    base = s.seq[i].upper()
                    col_dict[base].append(s.id)
                    if base in bases:
                        col_counter[base]+=1
                    
                majority = col_counter.most_common(1)[0][0] # majority base at this position

                for base in col_dict:
                    if len(col_dict[base]) == 1:
                        unique_mutations[col_dict[base][0]].add(i) # if only one sequence has this base, it's a unique mutation
                
                # if the snp is within a couple bases of an N, may be an issue with coverage/ alignment
                for s in aln:
                    # if the variant itself isn't N or a gap and isn't the majority base
                    if s.seq[i] != "N" and s.seq[i] != majority and s.seq[i] != "-":
                        if "N" in s.seq[i-n_window:i+n_window+1]: # check a window around the snp for Ns 
                            snps_near_n[s.id].add(i)

                for s in aln:
                    if s.seq[i] != "N" and s.seq[i] != majority and s.seq[i] != "-":
                        if "-" in s.seq[i-gap_window:i+gap_window+1]:# check a window around the snp for gaps
                            snps_near_gap[s.id].add(i)

        # identify clustered snps (e.g. x or more unique snps within y bases)
        clustered_sites = set()
        for s in aln:
            unique = unique_mutations[s.id]
            s_unique = sorted(unique)
            for i,val in enumerate(s_unique):

                if len(s_unique) > i+snp_count-1: # if there are enough snps remaining to meet the threshold
                    
                    snps_in_window = s_unique[i:i+snp_count] # get the next snp_count snps
                    # if x snps are within y bases
                    if snps_in_window[-1] < val+snp_window: 
                        for snp in snps_in_window:
                            clustered_snps[s.id].add(snp)

    return unique_mutations, snps_near_n, snps_near_gap, clustered_snps


def run_alignment_qc(
    alignment_path,
    outdir,
    genbank_path=None,
    reference_id=None,
    n_threshold=DEFAULT_N_THRESHOLD,
    n_window=2,
    gap_window=1,
    cluster_window=DEFAULT_CLUSTER_WINDOW,
    cluster_count=DEFAULT_CLUSTER_COUNT,
    mask_clustered=True,
    mask_n_adjacent=True,
    mask_gap_adjacent=True,
    mask_frame_break=True,
):
    """Run a series of alignment QC checks and write summary outputs.

    Returns a dict summary with details.
    """
    aln = read_alignment(alignment_path)

    summary = {}
    flagged_n = find_high_N_sequences(aln, threshold=n_threshold)
    summary['high_n_sequences'] = flagged_n

    unique_mutations, snps_near_n, snps_near_gap, clustered_snps = analyze_alignment(
        aln, n_window=n_window, gap_window=gap_window, snp_window=cluster_window, snp_count=cluster_count
    )
    
    frame_sites = {}
    if mask_frame_break and genbank_path:
        try:
            frame_sites = find_frame_breaking_indels(aln, genbank_path, reference_id=reference_id)
        except Exception as exc:
            logging.warning("Frame-breaking indel check skipped due to error: %s", exc)

    def _add_site(merged_sites, pos, seq_id, note):
        if pos not in merged_sites:
            merged_sites[pos] = {
                KEY_NAME: pos,
                KEY_MINIMUM: pos,
                KEY_MAXIMUM: pos,
                KEY_LENGTH: 1,
                KEY_PRESENT_IN: [seq_id],
                KEY_NOTE: set([note])
            }
            return
        if seq_id not in merged_sites[pos][KEY_PRESENT_IN]:
            merged_sites[pos][KEY_PRESENT_IN].append(seq_id)
        merged_sites[pos][KEY_NOTE].add(note)

    # build mask-site dict
    merged = {}
    if mask_n_adjacent:
        for seq_id, sites in snps_near_n.items():
            for site in sites:
                _add_site(merged, site + 1, seq_id, NOTE_N_ADJACENT)

    if mask_gap_adjacent:
        for seq_id, sites in snps_near_gap.items():
            for site in sites:
                _add_site(merged, site + 1, seq_id, NOTE_GAP_ADJACENT)

    if mask_clustered:
        for seq_id, sites in clustered_snps.items():
            for site in sites:
                _add_site(merged, site + 1, seq_id, NOTE_CLUSTERED_SNPS)

    # merge frame-breaking sites
    for site, info in frame_sites.items():
        if site not in merged:
            merged[site] = info
        else:
            merged[site][KEY_PRESENT_IN].extend(info[KEY_PRESENT_IN])
            merged[site][KEY_NOTE].update(info[KEY_NOTE])

    aln_len = aln.get_alignment_length()
    sequence_flag_counts = collections.Counter()
    sequence_flag_sites = collections.defaultdict(set)
    for site, info in merged.items():
        for seq_id in info[KEY_PRESENT_IN]:
            sequence_flag_counts[seq_id] += 1
            sequence_flag_sites[seq_id].add(site)

    flagged_sequences = {
        seq_id for seq_id, count in sequence_flag_counts.items() if count > 20
    }

    mask_file = os.path.join(outdir, 'mask_sites.csv')
    with open(mask_file, 'w') as fw:
        writer = csv.DictWriter(
            fw,
            lineterminator='\n',
            fieldnames=["flagged", "type", "Minimum", "Maximum", "Length", "present_in", "note"],
        )
        writer.writeheader()

        for seq_id in sorted(flagged_sequences):
            sites = sorted(sequence_flag_sites.get(seq_id, []))
            writer.writerow({
                "flagged": seq_id,
                "type": "sequence_record",
                "Minimum": 1,
                "Maximum": aln_len,
                "Length": aln_len,
                "present_in": seq_id,
                "note": ",".join(str(site) for site in sites),
            })

        for site in sorted(merged):
            row = merged[site]
            present_in = [seq_id for seq_id in row[KEY_PRESENT_IN] if seq_id not in flagged_sequences]
            if not present_in:
                continue
            note = ';'.join(sorted(row[KEY_NOTE]))
            if len(present_in) > 10:
                present_in_val = 'many'
            else:
                present_in_val = ';'.join(present_in)
            writer.writerow({
                "flagged": site,
                "type": "site",
                "Minimum": site,
                "Maximum": site,
                "Length": 1,
                "present_in": present_in_val,
                "note": note,
            })

    summary[KEY_SITES_TO_MASK] = merged
    summary[KEY_MASK_FILE] = mask_file
    # determine overall alignment health
    summary[KEY_ISSUES_FOUND] = bool(flagged_n or merged)
    return summary


